import logging

import confluent_kafka

import bspump.asab as asab

from ..abc.sink import Sink

#

L = logging.getLogger(__name__)

#


class KafkaSink(Sink):
    """
            Description: KafkaSink is a sink processor that forwards the event to a Apache Kafka specified by a KafkaConnection object.

            KafkaSink expects bytes as an input. If the input is string or dictionary, it is automatically transformed to bytes
            using encoding charset specified in the configuration.

    .. code:: python

            class KafkaPipeline(bspump.Pipeline):

                            def __init__(self, app, pipeline_id):
                                            super().__init__(app, pipeline_id)
                                            self.build(
                                                            bspump.kafka.KafkaSource(app, self, "KafkaConnection", config={'topic': 'messages'}),
                                                            bspump.kafka.KafkaSink(app, self, "KafkaConnection", config={'topic': 'messages2'}),
                            )

            There are two ways to use KafkaSink:
            - Specify a single topic in KafkaSink config - topic, to be used for all the events in pipeline.
            - Specify topic separetly for each event in event context - context['kafka_topic'].
                            Topic from configuration is than used as a default topic.
                            To provide business logic for event distribution, you can create topic selector processor.
            Processor example:

    .. code:: python

            class KafkaTopicSelector(bspump.Processor):

                            def process(self, context, event):
                                            if event.get("weight") > 10:
                                                            context["kafka_topic"] = "heavy"
                                            else:
                                                            context["kafka_topic"] = "light"

                                            return event

            Every kafka message can be a key:value pair. Key is read from event context - context['kafka_key'].
            If kafka_key is not provided, key defaults to None.

            Standard Kafka configuration options can be used,
            as specified in librdkafka library,
            where the options are simply passed to:

            https://github.com/edenhill/librdkafka/blob/master/CONFIGURATION.md
    """

    ConfigDefaults = {
        "topic": "unconfigured",
        "watermark.low": "40000",
        "watermark.high": "90000",
        "batch.num.messages": "100000",
        "linger.ms": "500",  # This settings makes a significant impact on the throughtput
        "batch.size": "1000000",
        "poll.timeout": "0.2",
        # "compression.type": "snappy",
    }

    def __init__(self, app, pipeline, connection, id=None, config=None):
        super().__init__(app, pipeline, id=id, config=config)

        connection = pipeline.locate_connection(app, connection)

        producer_config = {}

        # Copy connection options
        for key, value in connection.Config.items():
            producer_config[key.replace("_", ".")] = value

        # Copy configuration options, avoid the topic, watermark and poll.timeout params
        for key, value in self.Config.items():
            if key == "topic" or key.startswith("watermark") or key == "poll.timeout":
                continue

            producer_config[key.replace("_", ".")] = value

        self.Producer = confluent_kafka.Producer(producer_config, logger=L)

        self.Topic = self.Config["topic"]
        self.LowWatermark = int(self.Config["watermark.low"])
        self.HighWatermark = int(self.Config["watermark.high"])
        self.PollTimeout = float(self.Config["poll.timeout"])
        self.IsThrottling = False

        self.ProactorService = self.Pipeline.App.get_service("asab.ProactorService")

        app.PubSub.subscribe_all(self)

    @asab.subscribe("Application.tick!")
    async def _on_tick(self, event_name):
        if self.IsThrottling and (len(self.Producer) < self.LowWatermark):
            self.IsThrottling = False
            self.Pipeline.throttle(self, False)

        await self.ProactorService.execute(self.Producer.flush, self.PollTimeout)

    def process(self, context, event: bytes):
        try:
            self.Producer.produce(
                context["kafka_topic"] if "kafka_topic" in context else self.Topic,
                value=event,
                key=context["kafka_key"] if "kafka_key" in context else None,
                headers=(
                    context["kafka_headers"] if "kafka_headers" in context else None
                ),
            )

        except Exception as e:
            L.exception("Error occurred when sending data to Kafka: '{}'".format(e))

        if not self.IsThrottling and (len(self.Producer) > self.HighWatermark):
            self.IsThrottling = True
            self.Pipeline.throttle(self, True)



import logging

import motor.motor_asyncio
import pymongo

from ..abc.connection import Connection

#

L = logging.getLogger(__name__)

#


class MongoDBConnection(Connection):
    """
    Examples of configurations:

    [connection:Mongo]
    host=localhost
    port=27017


    [connection:Mongo]
    host=mongodb://localhost:27017

    [connection:Mongo]
    host=mongodb://host1,host2/?replicaSet=my-replicaset-name
    """

    ConfigDefaults = {
        "host": "localhost",  # hostname or IP address or Unix domain socket path of a single mongod or mongos instance to connect to, or a mongodb URI, or a list of hostnames / mongodb URIs.
        "port": 27017,
        "username": "",
        "password": "",
        "max_pool_size": 100,
        "min_pool_size": 0,
        "max_idle_time": "",
        "socket_timeout": "",
        "connect_timeout": "",
        "server_selection_timeout": "",
        "wait_queue_timeout": "",
        "wait_queue_multiple": "",
        "heartbeat_frequency": 10 * 1000,  # 10 seconds
        "database": "database",
    }

    def __init__(self, app, id=None, config=None):
        super().__init__(app, id=id, config=config)

        # TODO: SSL ...

        self.Client = motor.motor_asyncio.AsyncIOMotorClient(
            host=self.Config["host"],
            port=_get_int_or_none(self.Config, "port"),
            username=_get_str_or_none(self.Config, "username"),
            password=_get_str_or_none(self.Config, "password"),
            maxPoolSize=_get_int_or_none(self.Config, "max_pool_size"),
            minPoolSize=_get_int_or_none(self.Config, "min_pool_size"),
            maxIdleTimeMS=_get_int_or_none(self.Config, "max_idle_time"),
            socketTimeoutMS=_get_int_or_none(self.Config, "socket_timeout"),
            connectTimeoutMS=_get_int_or_none(self.Config, "connect_timeout"),
            waitQueueTimeoutMS=_get_int_or_none(self.Config, "wait_queue_timeout"),
            waitQueueMultiple=_get_int_or_none(self.Config, "wait_queue_multiple"),
            heartbeatFrequencyMS=_get_int_or_none(self.Config, "heartbeat_frequency"),
            appname=id,
            driver=pymongo.driver_info.DriverInfo(
                name="bspump.MongoDBConnection",
                # TODO: version=...
                platform="BitSwan",
            ),
            io_loop=app.Loop,
        )

        self.Database = self.Config["database"]


# TODO: Generalize this function to Configurable
def _get_int_or_none(config_obj, key):
    v = config_obj.get(key)
    if isinstance(v, str) and len(v) == 0:
        return None
    try:
        return int(v)
    except ValueError:
        L.error("Expects integer value", struct_data={"key": key, "value": v})
        return None


# TODO: Generalize this function to Configurable
def _get_str_or_none(config_obj, key):
    v = config_obj.get(key)
    if len(v) == 0:
        return None
    return v



from .base_field import BaseField
from .field import Field
from .float_field import FloatField
from .int_field import IntField
from .checkbox_field import CheckboxField
from .text_field import TextField
from .choice_field import ChoiceField
from .field_set import FieldSet
from .file_field import FileField
from .raw_json import RawJSONField
from .button import Button
from .date_field import DateField
from .datetime_field import DateTimeField
from .table import TableComponent

__all__ = [
    "BaseField",
    "Field",
    "FloatField",
    "IntField",
    "CheckboxField",
    "TextField",
    "ChoiceField",
    "FieldSet",
    "RawJSONField",
    "FileField",
    "Button",
    "DateField",
    "DateTimeField",
    "TableComponent",
]



import logging
import queue

import aiohttp.web
import itertools

from ...abc.sink import Sink

#

L = logging.getLogger(__name__)

#


class WebServiceStreamResponse(aiohttp.web.StreamResponse):
    Counter = itertools.count(1, 1)

    def __init__(self, sink):
        super().__init__()
        self.Id = "{}.{}#{}".format(sink.Pipeline.Id, sink.Id, next(self.Counter))
        self.Sink = sink
        self.Queue = queue.Queue()

    # Context Manager

    async def __aenter__(self):
        self.Sink.Responses[self.Id] = self

    async def __aexit__(self, exc_type, exc, tb):
        # Remove myself from a sink
        del self.Sink.Responses[self.Id]
        self.Sink = None

        await self.flush_events()
        await self.write_eof()

    # Queue management

    def put_event(self, event):
        self.Queue.put_nowait(event)

    async def flush_events(self):
        while not self.Queue.empty():
            event = self.Queue.get_nowait()
            await self.write(event)


class WebServiceSink(Sink):
    """
    Example of use:

            async def endpoint(self, request):
                    response = await self.WebServiceSink.response(request)
                    async with response:
                            data = await request.read()
                            await self.WebServiceSource.put({self.WebServiceSink.CONTEXT_RESPONSE_ID:response.Id}, data, request)
                    return response

    Note: Source doesn't necessarily needs to originate at WebServiceSink, it could be e.g. any triggered source.

    """

    CONTEXT_RESPONSE_ID = "webservicesink.response_id"
    CONTEXT_TYPE = "application/octet-stream"

    def __init__(self, app, pipeline, id=None, config=None):
        super().__init__(app, pipeline, id, config)
        self.Responses = {}
        app.PubSub.subscribe("Application.tick!", self.on_tick)

    def process(self, context, event):
        response_id = context.get(self.CONTEXT_RESPONSE_ID)
        if response_id is None:
            L.warning("Missing CONTEXT_RESPONSE_ID entry in the context")
            return

        response = self.Responses[response_id]
        if response_id is None:
            L.warning("Cannot locate web response '{}'".format(response_id))
            return

        response.put_event(event)

    async def response(self, request):
        response = WebServiceStreamResponse(self)
        response.content_type = self.CONTEXT_TYPE
        await response.prepare(request)
        return response

    async def on_tick(self, event_name):
        for response in self.Responses.values():
            await response.flush_events()



from aioresponses import aioresponses

import bspump.unittest
from bspump.influxdb import InfluxDBConnection, InfluxDBSink


class TestInfluxDBSink(bspump.unittest.ProcessorTestCase):
    @aioresponses()
    def test_influxdb_sink(self, mocked):
        mocked.post("http://test:8086/write?db=test", status=204)

        svc = self.App.get_service("bspump.PumpService")
        svc.add_connection(
            InfluxDBConnection(
                self.App,
                "InfluxDBConnection",
                config={
                    "output_bucket_max_size": "1",
                    "url": "http://test:8086/",
                    "db": "test",
                },
            )
        )

        events = [
            (None, ("collection", "tag1=a,tag2=b", "field=12", 1234567890)),
        ]

        self.set_up_processor(InfluxDBSink, connection="InfluxDBConnection")

        output = self.execute(events)

        requests = mocked.requests.items()
        self.assertEqual(1, len(requests))
        request = next(iter(requests))
        self.assertEqual("http://test:8086/write?db=test", str(request[0][1]))
        self.assertEqual(
            "collection,tag1=a,tag2=b field=12 1234567890000000000",
            request[1][0][1]["data"].strip(),
        )

        self.assertEqual(output, [])



import logging

from ..abc.source import TriggerSource


L = logging.getLogger(__name__)


class AnalyzingSource(TriggerSource):
    """
    The `AnalyzingSource` is triggered source, which expects `matrix_id` as an input.
    Each trigger fire it calls `analyze()` function of the `Matrix` and expects
    a complex event as an output. A complex event can be array of events, aggregation of events (average, max, min etc.)
    """

    def __init__(self, app, pipeline, matrix_id, id=None, config=None):
        """
        Description:

        **Parameters**

        app : Application
                Name of the Application

        pipeline : Pipeline
                Name of the Pipeline

        matrix_id : str
                ID of the matrix.

        id : str, default = None
                ID

        config : JSON, default = None
                configuration file containing additional information.


        """
        super().__init__(app, pipeline, id=id, config=config)
        svc = app.get_service("bspump.PumpService")
        self.AnalyzeMatrix = svc.locate_matrix(matrix_id)

    async def cycle(self):
        """
        Description:

        """
        event = await self.AnalyzeMatrix.analyze()
        await self.process(event)



from .connection import MongoDBConnection
from .source import MongoDBSource
from .lookup import MongoDBLookup
from .sink import MongoDBSink
from .changestreamsource import MongoDBChangeStreamSource

__all__ = (
    "MongoDBChangeStreamSource",
    "MongoDBConnection",
    "MongoDBLookup",
    "MongoDBSource",
    "MongoDBSink",
)



import datetime
import hashlib
import os
import os.path
import logging

import aiohttp.web

from bspump.asab.web.rest import json_response
from bspump.asab.web.auth import noauth

from ..__version__ import __build__ as bspump_build
from ..__version__ import __version__ as bspump_version

#

L = logging.getLogger(__name__)

#


@noauth
async def pipelines(request):
    app = request.app["app"]
    svc = app.get_service("bspump.PumpService")
    return json_response(request, svc.Pipelines)


@noauth
async def example_trigger(request):
    app = request.app["app"]
    app.PubSub.publish("mymessage!")
    return json_response(request, {"ok": 1})


@noauth
async def example_internal(request):
    app = request.app["app"]
    svc = app.get_service("bspump.PumpService")
    source = svc.locate("SampleInternalPipeline.*InternalSource")
    source.put({"event": "example"})
    return json_response(request, {"ok": 1})


@noauth
async def lookup_list(request):
    app = request.app["app"]
    svc = app.get_service("bspump.PumpService")
    return json_response(
        request, [lookup.rest_get() for lookup in svc.Lookups.values()]
    )


@noauth
async def lookup_meta(request):
    lookup_id = request.match_info.get("lookup_id")
    app = request.app["app"]
    svc = app.get_service("bspump.PumpService")
    try:
        lookup = svc.locate_lookup(lookup_id)
    except KeyError:
        raise aiohttp.web.HTTPNotFound()
    return json_response(request, lookup.rest_get())


@noauth
async def lookup(request):
    lookup_id = request.match_info.get("lookup_id")
    app = request.app["app"]
    svc = app.get_service("bspump.PumpService")
    request_etag = request.headers.get("ETag")

    try:
        lookup = svc.locate_lookup(lookup_id)
    except KeyError:
        raise aiohttp.web.HTTPNotFound()

    try:
        data = lookup.serialize()
    except AttributeError:
        raise aiohttp.web.HTTPNotImplemented()

    assert isinstance(data, bytes)

    response_etag = hashlib.sha1(data).hexdigest()
    if (request_etag is not None) and (request_etag == response_etag):
        raise aiohttp.web.HTTPNotModified()

    return aiohttp.web.Response(
        body=data,
        status=200,
        headers={"ETag": response_etag},
        content_type="application/octet-stream",
    )


@noauth
async def manifest(request):
    """
            $ curl http://localhost:8080/manifest?pretty=true
            {
                    "ASAB_VERSION": "18.12b1",
                    "DOCKER_HOST": "${DOCKER_HOST}",
                    "BSPUMP_VERSION": "1.2.3",
                    "APP_VERSION": "1.2.3",
                    "MANIFEST_MTIME": "2019-01-29T21:54:15.780668"
            }

            Served from `/manifest` file.
            The format of the file is:
    KEY1=VALUE
    KEY2=${ENVIRONMENT_VARIABLE}
    """

    app = request.app["app"]

    d = {
        "BSPUMP_VERSION": bspump_version,
        "BSPUMP_BUILD": bspump_build,
        "LAUNCHED_AT": datetime.datetime.utcfromtimestamp(app.LaunchTime).isoformat(),
    }

    container_host = os.environ.get("CONTAINER_HOST")
    if container_host is not None:
        d["CONTAINER_HOST"] = container_host

    try:
        fname = "/manifest"
        mtime = os.path.getmtime(fname)
        with open(fname) as f:
            for line in f:
                k, v = line.strip().split("=", 1)
                if k not in d:
                    d[k] = os.path.expandvars(v)
        d["MANIFEST_MTIME"] = datetime.datetime.utcfromtimestamp(mtime).isoformat()
    except FileNotFoundError:
        pass
    return json_response(request, d)


def initialize_web(container):
    # Add routes

    # API VERSION 1
    container.WebApp.router.add_get("/bspump/v1/pipelines", pipelines)
    container.WebApp.router.add_get("/bspump/v1/example/trigger", example_trigger)
    container.WebApp.router.add_get("/bspump/v1/example/internal", example_internal)

    container.WebApp.router.add_get("/bspump/v1/lookup", lookup_list)
    container.WebApp.router.add_get("/bspump/v1/lookup/{lookup_id}", lookup)
    container.WebApp.router.add_get("/bspump/v1/lookup/{lookup_id}/meta", lookup_meta)

    container.WebApp.router.add_get("/bspump/v1/manifest", manifest)

    return container



from aiohttp.web import Request

from bspump.http.web.components.field import Field
from bspump.http.web.template_env import env


class CheckboxField(Field):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.default = self.default or False

    def inner_html(self, default="", readonly=False):
        readonly_attr = "disabled" if readonly else ""

        template = env.get_template("checkbox-field.html")
        return template.render(
            default=default,
            readonly=readonly,
            default_classes=self.default_classes,
            default_input_props=self.default_input_props,
            readonly_attr=readonly_attr,
        )

    def clean(self, data, request: Request | None = None):
        if type(data.get(self.name)) == str:
            data[self.name] = data.get(self.name, False) == "on"



class Anomaly(dict):
    """
    Description: 	Anomaly is an abstract class to be overriden for a specific anomaly and its type.

    :return:

    Implement: TYPE, on_tick

    |

    """

    TYPE = None

    def is_closed(self):
        """
        Description:

        :return: sets status to closed

        |

        """
        return self.get("status") == "closed"

    def close(self, current_time):
        """
        Description:

        """
        self["ts_end"] = current_time
        self["D"] = self["ts_end"] - self["@timestamp"]
        self["status"] = "closed"

    async def on_tick(self, current_time):
        """
        Description:

        :hint: Implement to perform operations on the anomaly, f. e. close.
        """
        raise NotImplementedError()



from ..abc.sink import Sink


class SmtpSink(Sink):
    def __init__(self, app, pipeline, connection, id=None, config=None):
        super().__init__(app, pipeline, id=id, config=config)

        self.Connection = pipeline.locate_connection(app, connection)

        app.PubSub.subscribe("SMTPConnection.pause!", self._connection_throttle)
        app.PubSub.subscribe("SMTPConnection.unpause!", self._connection_throttle)

    def _connection_throttle(self, event_name, connection):
        if connection != self.Connection:
            return

        if event_name == "SMTPConnection.pause!":
            self.Pipeline.throttle(self, True)
        elif event_name == "SMTPConnection.unpause!":
            self.Pipeline.throttle(self, False)
        else:
            raise RuntimeError("Unexpected event name '{}'".format(event_name))

    def process(self, context, event):
        self.Connection.consume(event)



import typing
from .service import StorageServiceABC
from .upsertor import UpsertorABC
from .exceptions import DuplicateError


class InMemoryUpsertor(UpsertorABC):
    def __init__(self, storage, collection, obj_id, version=None):
        super().__init__(storage, collection, obj_id, version)
        if self.ObjId is None:
            # generate a random unique binary ID
            self.ObjId = self.generate_id()

    async def execute(
        self,
        custom_data: typing.Optional[dict] = None,
        event_type: typing.Optional[str] = None,
    ) -> typing.Union[str, bytes]:
        """Commit the changes prepared in upsertor.

        :custom_data (dict, optional): Not implemented yet. Defaults to None.
        :event_type (str, optional): Not implemented yet. Defaults to None.

        Raises: :RuntimeError: Raised if the object ID was not found in the previous version.

        Returns:
                :str | bytes: ID of the created or updated document.
        """

        # TODO: Implement webhook call
        id_name = self.get_id_name()

        # Get the object
        if self.Version == 0:
            obj = {id_name: self.ObjId}
            self.Storage._set(self.Collection, self.ObjId, obj)

        else:
            obj = await self.Storage.get(self.Collection, self.ObjId)
            if obj is None:
                if self.Version is None:
                    obj = {id_name: self.ObjId}
                    self.Storage._set(self.Collection, self.ObjId, obj)
                else:
                    raise RuntimeError(
                        "Previous version of '{}' not found".format(self.ObjId)
                    )

        for k, v in self.ModSet.items():
            obj[k] = v

        for k, v in self.ModUnset.items():
            obj.pop(k, None)

        for k, v in self.ModInc.items():
            o = obj.pop(k, 0)
            obj[k] = o + v

        for k, v in self.ModPush.items():
            o = obj.pop(k, None)
            if o is None:
                o = list()
            o.extend(v)
            obj[k] = o

        for k, v in self.ModPull.items():
            o = obj.pop(k, None)
            if o is None:
                o = list()
            for x in v:
                try:
                    o.remove(x)
                except ValueError:
                    pass
            obj[k] = o

        return self.ObjId


class StorageService(StorageServiceABC):
    def __init__(self, app, service_name):
        super().__init__(app, service_name)
        self.InMemoryCollections = {}

    def upsertor(self, collection: str, obj_id=None, version=0) -> InMemoryUpsertor:
        """Obtain an in-memory upsertor for given collection and possibly for the specified object.

        :collection (str): The name of the collection.
        :obj_id (_type_, optional): The ID of the document to retrieve. Defaults to None.
        :version (int, optional): The version of the collection. Defaults to 0.

        Returns:
                :InMemoryUpsertor: Upsertor for given collection.

        """
        return InMemoryUpsertor(self, collection, obj_id, version)

    async def get(
        self, collection: str, obj_id: typing.Union[str, bytes], decrypt=None
    ) -> dict:
        """Retrieve a document from an in-memory collection by its ID.

        :collection (str): The name of the collection to retrieve the document from.
        :obj_id (str | bytes): The ID of the document to retrieve.
        :decrypt (_type_, optional): A list of field names to decrypt. Defaults to None.

        Returns:
                :dict: A dictionary representing the retrieved document.
                If `decrypt` is not None, the specified fields in the document are decrypted using AES decryption algorithm.

        """
        coll = self.InMemoryCollections[collection]
        data = coll[obj_id]
        if decrypt is not None:
            for field in decrypt:
                if field in data:
                    data[field] = self.aes_decrypt(data[field])
        return data

    async def get_by(self, collection: str, key: str, value, decrypt=None) -> dict:
        """
        Retrieve a document from an in-memory collection by key and value. Not implemented yet.

        Raises:
                :NotImplementedError: Not implemented on InMemoryStorage
        """
        raise NotImplementedError()

    async def delete(self, collection: str, obj_id):
        """
        Delete a document from an in-memory collection.

        :param collection: Collection to delete from
        :param obj_id: Object identification

        Raises:
                :KeyError: If `obj_id` not found in `collection`
        """
        coll = self.InMemoryCollections[collection]
        del coll[obj_id]

    def _set(self, collection: str, obj_id, obj):
        try:
            coll = self.InMemoryCollections[collection]
        except KeyError:
            coll = {}
            self.InMemoryCollections[collection] = coll

        nobj = coll.setdefault(obj_id, obj)
        if nobj != obj:
            raise DuplicateError("Already exists", obj_id)



from unittest.mock import patch, call

import bspump.file
import bspump.unittest


class CustomFileBlockSink(bspump.file.FileBlockSink):
    ConfigDefaults = {"path": "/path/doesn't/exists.csv"}


class TestFileBlockSink(bspump.unittest.ProcessorTestCase):
    @patch("bspump.file.fileblocksink.os.open")
    def test_file_block_sink(self, mocked_os_open):
        mocked_os_open.return_value = 3  # File descriptor
        events = [
            (None, b"deadc0de"),
        ]

        self.set_up_processor(CustomFileBlockSink)
        with patch("bspump.file.fileblocksink.os.fdopen") as mock_file:
            output = self.execute(events)
            self.assertIn(call().__enter__().write(b"deadc0de"), mock_file.mock_calls)
            self.assertEqual(mock_file.call_args, call(3, "wb"))
            self.assertEqual(output, [])



import collections.abc

from ..abc.processor import Processor


class MappingTransformator(Processor):
    """
    Description:

    |

    """

    def __init__(self, app, pipeline, id=None, config=None):
        """
        Description:

        |

        """
        super().__init__(app, pipeline, id=id, config=config)
        self.Mapping = self.build(app)
        self.Default = None

    def build(self, app):
        """
        Description:

        |

        """
        raise NotImplementedError()

    def _map(self, item):
        """
        Description:

        :return: ??

        |

        """
        key, value = item
        t = self.Mapping.get(key)
        if t is not None:
            return t(key, value)
        elif self.Default is not None:
            return self.Default(key, value)
        else:
            return key, value

    def process(self, context, event: collections.abc.Mapping) -> dict:
        """
        Description:

        :return: dict(map(self._map, event.items()))

        |

        """
        return dict(map(self._map, event.items()))



from ..abc.processor import Processor
from .builder import ExpressionBuilder
from .optimizer import ExpressionOptimizer


class DeclarativeProcessor(Processor):
    @classmethod
    def construct(cls, app, pipeline, definition: dict):
        _id = definition.get("id")
        config = definition.get("config")
        declaration = definition.get("declaration")
        return cls(app, pipeline, declaration=declaration, id=_id, config=config)

    def __init__(self, app, pipeline, declaration, library=None, id=None, config=None):
        super().__init__(app, pipeline, id=id, config=config)
        self.Declaration = declaration
        self.Builder = ExpressionBuilder(app, library)
        self.ExpressionOptimizer = ExpressionOptimizer(app)
        self.Expressions = None

    async def initialize(self):
        expressions = await self.Builder.parse(self.Declaration)
        self.Expressions = self.ExpressionOptimizer.optimize_many(expressions)

    def process(self, context, event):
        for expression in self.Expressions:
            event = expression(context, event)
            if event is None:
                return None

        if not event:
            return None

        return event



import time
from ...abc import Expression


class NOW(Expression):
    """
    Unit test usage to mock current time:

            import time
            import unittest.mock

            class MetaTestCase(unittest.TestCase):

                    @unittest.mock.patch('bspump.declarative.expression.NOW.Time')
                    def runTest(self, mock_time):
                            mock_time.return_value = 1234567890 # Mocked timestamp
                            ...
    """

    Attributes = {}
    Time = time.time

    Category = "Date/Time"

    def __init__(self, app, *, value):
        super().__init__(app)
        assert value == ""

    def __call__(self, context, event, *args, **kwargs):
        return self.Time()

    def get_outlet_type(self):
        return float.__name__



import bspump.asab as asab

from .processor import DeclarativeProcessor
from .generator import DeclarativeGenerator
from .timewindowanalyzer import DeclarativeTimeWindowAnalyzer

from .builder import ExpressionBuilder
from .optimizer import ExpressionOptimizer
from .declerror import DeclarationError
from .segmentbuilder import SegmentBuilder

from .abc import Expression
from .abc import SequenceExpression

from .dot import declaration_to_dot, declaration_to_dot_stream


asab.Config.add_defaults(
    {
        "declarations": {
            "timezone": ""  # Default timezone to be used by DATETIME expression, such as Europe/Prague
        }
    }
)


__all__ = [
    "DeclarativeProcessor",
    "DeclarativeGenerator",
    "DeclarativeTimeWindowAnalyzer",
    "ExpressionBuilder",
    "DeclarationError",
    "SegmentBuilder",
    "ExpressionOptimizer",
    "Expression",
    "SequenceExpression",
    "declaration_to_dot",
    "declaration_to_dot_stream",
]



import asyncio
import logging

import pika
import pkg_resources

from ..abc.source import Source


L = logging.getLogger(__name__)


class AMQPSource(Source):
    """
    Description:


    """

    ConfigDefaults = {
        "queue": "q",
        "error_exchange": "error",
        "prefetch_count": "1000",
    }

    def __init__(self, app, pipeline, connection, id=None, config=None):
        super().__init__(app, pipeline, id=id, config=config)

        self._connection = pipeline.locate_connection(app, connection)
        self._channel = None
        self._channel_ready = asyncio.Event()
        self._channel_ready.clear()
        self._error_exchange = self.Config["error_exchange"]
        self._queue = asyncio.Queue()

        self._connection.PubSub.subscribe(
            "AMQPConnection.open!", self._on_connection_open
        )
        self._connection.PubSub.subscribe(
            "AMQPConnection.close!", self._on_connection_close
        )

    async def main(self):
        """
        Description:


        """

        if self._connection.ConnectionEvent.is_set() and self._channel is None:
            self._on_connection_open(".local!")

        try:
            while 1:
                await self._channel_ready.wait()
                await self.Pipeline.ready()
                method, properties, body = await self._queue.get()
                await self.process_message(method, properties, body)
                self._channel.basic_ack(method.delivery_tag)

        except asyncio.CancelledError:
            pass

        except BaseException as e:
            L.exception("Error when processing AMQP message")
            self.Pipeline.set_error(None, None, e)

        # Requeue rest of the messages
        while not self._queue.empty():
            method, properties, body = await self._queue.get()
            self._channel.basic_nack(method.delivery_tag, requeue=True)

        if self._channel is not None:
            self._channel.close()
            self._channel = None
            self._channel_ready.clear()

    async def process_message(self, method, properties, body):
        """
        Description:


        """
        context = {"amqp:method": method, "amqp:properties": properties}
        await self.process(body, context=context)

    def _on_connection_open(self, event_name):
        assert self._channel is None
        self._channel = self._connection.Connection.channel(
            on_open_callback=self._on_channel_open
        )
        self._channel_ready.set()

    def _on_connection_close(self, event_name):
        self._channel = None
        self._channel_ready.clear()

    def _on_qos_applied(self, channel):
        if pkg_resources.parse_version(pika.__version__) >= pkg_resources.parse_version(
            "1.0.a"
        ):
            self._channel.basic_consume(self.Config["queue"], self._on_consume_message)
        else:
            self._channel.basic_consume(self._on_consume_message, self.Config["queue"])

    def _on_channel_open(self, channel):
        channel.basic_qos(
            callback=self._on_qos_applied,
            prefetch_count=int(self.Config["prefetch_count"]),
        )

    def _on_consume_message(self, channel, method, properties, body):
        try:
            self._queue.put_nowait((method, properties, body))
        except Exception:
            channel.basic_nack(method.delivery_tag, requeue=True)

    @classmethod
    def construct(cls, app, pipeline, definition: dict):
        """
        Description:


        """
        newid = definition.get("id")
        config = definition.get("config")
        connection = definition["args"]["connection"]
        return cls(app, pipeline, connection, newid, config)


class AMQPFullMessageSource(AMQPSource):
    """
    Description:


    """

    def process_message(self, method, properties, body):
        """
        Description:


        """
        self.process((method, properties, body))



import logging

from ..abc.processor import Processor

###

L = logging.getLogger(__name__)

###


class AttributeFilter(Processor):
    """
    This is processor implenting a simple attribute filter.
    If 'inclusive' is False, all fields from event matching with lookup will be deleted,
    otherwise all fields not from lookup will be deleted from event.
    """

    def __init__(
        self, app, pipeline, lookup=None, inclusive=False, id=None, config=None
    ):
        super().__init__(app, pipeline, id=id, config=config)
        self.Inclusive = inclusive

        # Lookups discovery
        if lookup is not None:
            svc = app.get_service("bspump.PumpService")
            self.Lookup = svc.locate_lookup(lookup)
        else:
            self.Lookup = None

    def predicate(self, event):
        return True

    def get_fields(self, event):
        return set()

    def filter_fields(self, event):
        fields = self.get_fields(event)
        for event_key in list(event):
            if (event_key in fields) != self.Inclusive:
                event.pop(event_key)
        return event

    def process(self, context, event):
        if not self.predicate(event):
            return event

        return self.filter_fields(event)



import ssl
import time
import logging
import datetime
import urllib.parse
import typing
import contextlib

import aiohttp
from bspump.asab import Config
import bspump.asab as asab

from .service import StorageServiceABC
from .upsertor import UpsertorABC
from ..tls import SSLContextBuilder

#

L = logging.getLogger(__name__)

#

Config.add_defaults(
    {
        "asab:storage": {
            # You may specify multiple ElasticSearch nodes by e.g. http://es01:9200,es02:9200,es03:9200/
            "elasticsearch_url": "",
            "elasticsearch_username": "",
            "elasticsearch_password": "",
            "elasticsearch_api_key": "",
            # make the operation visible to search directly, options: true, false, wait_for
            # see: https://www.elastic.co/guide/en/elasticsearch/reference/current/docs-index_.html
            "refresh": "true",
            "scroll_timeout": "1m",
            # For SSL options such as `cafile`, please refer to tls.py
        }
    }
)


class ElasticError(Exception):
    pass


class StorageService(StorageServiceABC):
    """
    StorageService for Elastic Search. Depends on `aiohttp` library.
    """

    def __init__(self, app, service_name, config_section_name="asab:storage"):
        super().__init__(app, service_name)

        self.Refresh = Config.get(config_section_name, "refresh", fallback="true")
        self.ScrollTimeout = Config.get(
            config_section_name, "scroll_timeout", fallback="1m"
        )

        # always check if there is a url in the old config section first
        url = Config.getmultiline(config_section_name, "elasticsearch_url", fallback="")
        if len(url) > 0:
            asab.LogObsolete.warning(
                "Do not configure elasticsearch connection in [asab:storage]. Please use [elasticsearch] section with url, username and password parameters."
            )
        elif len(url) == 0:
            url = Config.getmultiline("elasticsearch", "url", fallback="")

        self.ServerUrls = get_url_list(url)

        if len(self.ServerUrls) == 0:
            raise RuntimeError("No ElasticSearch URL has been provided.")

        # Authorization: username or API-key
        username = Config.get(config_section_name, "elasticsearch_username")
        if len(username) == 0:
            username = Config.get("elasticsearch", "username", fallback="")

        password = Config.get(config_section_name, "elasticsearch_password")
        if len(password) == 0:
            password = Config.get("elasticsearch", "password", fallback="")

        api_key = Config.get(config_section_name, "elasticsearch_api_key")
        if len(api_key) == 0:
            api_key = Config.get("elasticsearch", "api_key", fallback="")

        # Create headers for requests
        self.Headers = build_headers(username, password, api_key)

        # Build ssl context
        if self.ServerUrls[0].startswith("https://"):
            # check if [asab:storage] section has data for SSL or default to the [elasticsearch] section
            if section_has_ssl_option(config_section_name):
                self.SSLContextBuilder = SSLContextBuilder(
                    config_section_name=config_section_name
                )
            else:
                self.SSLContextBuilder = SSLContextBuilder(
                    config_section_name="elasticsearch"
                )
            self.SSLContext = self.SSLContextBuilder.build(ssl.PROTOCOL_TLS_CLIENT)
        else:
            self.SSLContext = None

    @contextlib.asynccontextmanager
    async def request(self, method, path, data=None, json=None):
        """
        This method can be used to do a custom call to ElasticSearch like so:

        async with self.request("GET", "cluster/_health") as resp:
                ...

        """
        async with aiohttp.ClientSession() as session:
            for n, url in enumerate(self.ServerUrls, 1):
                try:
                    async with session.request(
                        method=method,
                        url=url + path,
                        ssl=self.SSLContext,
                        headers=self.Headers,
                        data=data,
                        json=json,
                    ) as resp:
                        if resp.status == 401:
                            raise ConnectionRefusedError(
                                "Response code 401: Unauthorized. Provide authorization by specifying either user name and password or api key."
                            )

                        yield resp
                        return

                except aiohttp.client_exceptions.ClientConnectorError:
                    if n == len(self.ServerUrls):
                        raise ConnectionError("Failed to connect to '{}'.".format(url))
                    else:
                        L.warning(
                            "Failed to connect to '{}', iterating to another node".format(
                                url
                            )
                        )

    async def is_connected(self) -> bool:
        """
        Check if the service is connected to ElasticSearch cluster.

        Raises:
                ConnectionError: Connection failed.

        Returns:
                bool: True if the service is connected.
        """
        async with self.request("GET", "") as resp:
            if resp.status not in {200, 201}:
                resp = await resp.json()
                L.error(
                    "Failed to connect to ElasticSearch.",
                    struct_data={
                        "code": resp.get("status"),
                        "reason": resp.get("error", {}).get("reason"),
                    },
                )
                return False

            else:
                L.info(
                    "Connected to ElasticSearch.", struct_data={"urls": self.ServerUrls}
                )
                return True

    async def get(self, index: str, obj_id: str, decrypt=None) -> dict:
        """
        Get object by its index and object ID.

        Args:
                index (str): Index for the query.
                obj_id (str): ID of the object.
                decrypt (None): Not implemented yet. Defaults to None.

        Raises:
                NotImplementedError: Encryption and decryption has not yet been implemented for ECS.
                ConnectionError: Connection failed.
                ConnectionRefusedError: Authorization required.
                KeyError: Object with the ID does not exist.

        Returns:
                The query result.
        """
        if decrypt is not None:
            raise NotImplementedError(
                "AES encryption for ElasticSearch not implemented"
            )

        async with self.request("GET", "{}/_doc/{}".format(index, obj_id)) as resp:
            if resp.status not in {200, 201}:
                resp = await resp.json()
                raise ConnectionError(
                    "Failed to retrieve data from ElasticSearch. Got {}: {}".format(
                        resp.get("status"), resp.get("error", {}).get("reason")
                    )
                )

            else:
                obj = await resp.json()
                if not obj.get("found"):
                    raise KeyError("No existing object with ID {}".format(obj_id))
                ret = obj["_source"]
                ret["_v"] = obj["_version"]
                ret["_id"] = obj["_id"]
                return ret

    async def get_by(self, collection: str, key: str, value, decrypt=None):
        raise NotImplementedError("get_by")

    async def delete(self, index: str, _id=None) -> dict:
        """
        Delete an entire index or document from that index.

        Args:
                index: Index to delete.
                _id: If specified, only document with the ID is deleted.

        Raises:
                ConnectionRefusedError: Authorization required (status 401)
                KeyError: No existing object with ID
                ConnectionError: Unexpected status code
                Exception: ClientConnectorError

        Returns:
                The deleted document or message that the entire index was deleted.
        """

        if _id:
            path = "{}/_doc/{}?refresh={}".format(index, _id, self.Refresh)
        else:
            path = "{}".format(index)

        async with self.request("DELETE", path) as resp:
            if resp.status == 404:
                raise KeyError("No existing object with ID {}".format(_id))

            elif resp.status not in {200, 201}:
                raise ConnectionError(
                    "Failed to retrieve data from ElasticSearch. Got {}: {}".format(
                        resp.get("status"), resp.get("error", {}).get("reason")
                    )
                )

            else:
                json_response = await resp.json()

                if json_response.get("acknowledged", False):
                    return json_response
                assert json_response["result"] == "deleted", "Document was not deleted"
                return json_response

    async def mapping(self, index: str) -> dict:
        """
        Retrieve mapping definitions for one index.

        :param index: Specified index.
        :type index: str
        :raise Exception: Connection failed.

        Returns:
                dict: Mapping definitions for the index.
        """
        async with self.request("GET", "{}/_mapping".format(index)) as resp:
            if resp.status != 200:
                raise ElasticError(
                    "Unexpected response code: {}: '{}'".format(
                        resp.status, await resp.text()
                    )
                )
            return await resp.json()

    async def get_index_template(self, template_name: str) -> dict:
        """
        Retrieve ECS Index template for the given template name.

        :param template_name: The name of the ECS template to retrieve.
        :type template_name: str
        :raise Exception: Raised if connection to all server URLs fails.
        :return: ElasticSearch Index template.
        """
        async with self.request(
            "GET", "_index_template/{}?format=json".format(template_name)
        ) as resp:
            if resp.status != 200:
                raise ElasticError(
                    "Unexpected response code: {}: '{}'".format(
                        resp.status, await resp.text()
                    )
                )
            return await resp.json()

    async def put_index_template(self, template_name: str, template: dict) -> dict:
        """
        Create a new ECS index template.

                :param template_name: The name of ECS template.
                :param template: Body for the request.
                :return: JSON response.
                :raise Exception: Raised if connection to all server URLs fails.
        """
        async with self.request(
            "PUT", "_index_template/{}".format(template_name), json=template
        ) as resp:
            if resp.status != 200:
                raise ElasticError(
                    "Unexpected response code: {}: '{}'".format(
                        resp.status, await resp.text()
                    )
                )

            return await resp.json()

    async def reindex(self, previous_index, new_index):
        data = {
            "source": {
                "index": previous_index,
            },
            "dest": {
                "index": new_index,
            },
        }
        async with self.request("POST", "_reindex", json=data) as resp:
            if resp.status != 200:
                raise AssertionError(
                    "Unexpected response code when reindexing: {}, {}".format(
                        resp.status, await resp.text()
                    )
                )

            print("------> REINDEX ;=)")
            return await resp.json()

    def upsertor(self, index: str, obj_id=None, version: int = 0):
        return ElasticSearchUpsertor(self, index, obj_id, version)

    async def list(
        self,
        index: str,
        _from: int = 0,
        size: int = 10000,
        body: typing.Optional[dict] = None,
    ) -> dict:
        """List data matching the index.

        :param index: Specified index.
        :param _from:  Starting document offset. Defaults to 0.
        :type _from: int
        :param size: The number of hits to return. Defaults to 10000.
        :type size: int
        :param body: An optional request body. Defaults to None.
        :type body: dict

        :return: The query search result.
        :raise Exception: Raised if connection to all server URLs fails.

        """
        if body is None:
            body = {"query": {"bool": {"must": {"match_all": {}}}}}

        async with self.request(
            "GET", "{}/_search?size={}&from={}&version=true".format(index, size, _from)
        ) as resp:
            if resp.status != 200:
                raise ElasticError(
                    "Unexpected response code: {}: '{}'".format(
                        resp.status, await resp.text()
                    )
                )

            return await resp.json()

    async def count(self, index) -> int:
        """
        Get the number of matches for a given index.

        :param index: The specified index.
        :return: The number of matches for a given index.
        :raise Exception: Connection failed.
        """

        async with self.request("GET", "{}/_count".format(index)) as resp:
            if resp.status != 200:
                raise ElasticError(
                    "Unexpected response code: {}: '{}'".format(
                        resp.status, await resp.text()
                    )
                )

            return await resp.json()

    async def indices(self, search_string=None):
        """
        Return high-level information about indices in a cluster, including backing indices for data streams.

        :param search_string: A search string. Default to None.
        """

        async with self.request(
            "GET",
            "_cat/indices/{}?format=json".format(
                search_string if search_string is not None else "*"
            ),
        ) as resp:
            if resp.status != 200:
                raise ElasticError(
                    "Unexpected response code: {}: '{}'".format(
                        resp.status, await resp.text()
                    )
                )
            return await resp.json()

    async def empty_index(self, index, settings=None):
        """
        Create an empty ECS index.
        """
        # TODO: There is an option here to specify settings (e.g. shard number, replica number etc) and mappings here

        if settings is None:
            settings = {}

        async with self.request("PUT", index, json=settings) as resp:
            if resp.status != 200:
                raise ElasticError(
                    "Unexpected response code: {}: '{}'".format(
                        resp.status, await resp.text()
                    )
                )
            return await resp.json()

    async def put_policy(self, policy_name, settings=None):
        """
        Create a lifecycle policy.
        """

        if settings is None:
            settings = {}

        async with self.request(
            "PUT", "_ilm/policy/{}".format(policy_name), json=settings
        ) as resp:
            if resp.status != 200:
                raise ElasticError(
                    "Unexpected response code: {}: '{}'".format(
                        resp.status, await resp.text()
                    )
                )

            return await resp.json()

    async def policies(self):
        """
        Return high-level information about ILM policies in a cluster, including backing indices for data streams.

        :param search_string: A search string. Default to None.
        """
        async with self.request("GET", "_ilm/policy") as resp:
            if resp.status != 200:
                raise ElasticError(
                    "Unexpected response code: {}: '{}'".format(
                        resp.status, await resp.text()
                    )
                )

            return await resp.json()


class ElasticSearchUpsertor(UpsertorABC):
    def __init__(self, storage, collection, obj_id, version=None):
        super().__init__(storage, collection, obj_id, version)

        now = int(time.time())

        self.ModSet["_m"] = now

        if version == 0:
            self.ModSet["_c"] = now  # Set the creation timestamp

    @classmethod
    def generate_id(cls):
        raise NotImplementedError("generate_id")

    async def execute(
        self,
        custom_data: typing.Optional[dict] = None,
        event_type: typing.Optional[str] = None,
    ):
        # TODO: Implement webhook call
        if self.ObjId is None:
            return await self._insert_new_object()
        else:
            return await self._update_existing_object()

    async def _insert_new_object(self):
        upsert_data = {}

        if self.Version is None:
            self.Version = 0

        if len(self.ModSet) > 0:
            for k, v in self.ModSet.items():
                upsert_data[k] = serialize(self.ModSet[k])

        if len(self.ModInc) > 0:
            # addobj['$inc'] = self.ModInc
            # raise NotImplementedError("yet")
            pass

        if len(self.ModPush) > 0:
            # addobj['$push'] = {k: {'$each': v} for k, v in self.ModPush.items()}
            raise NotImplementedError("yet")

        # This is insert of the new document, the ObjId is to be generated by the ElasicSearch
        async with self.Storage.request(
            "POST",
            "{}/_doc?refresh={}".format(self.Collection, self.Storage.Refresh),
            json=upsert_data,
        ) as resp:
            if resp.status not in {200, 201}:
                raise ConnectionError(
                    "Unexpected response code: {}".format(resp.status)
                )
            else:
                resp_json = await resp.json()
                self.ObjId = resp_json["_id"]
                return self.ObjId

    async def _update_existing_object(self):
        upsert_data = {"doc": {}, "doc_as_upsert": True}

        if len(self.ModSet) == 0:
            return

        for k, v in self.ModSet.items():
            upsert_data["doc"][k] = serialize(self.ModSet[k])

        async with self.Storage.request(
            "POST",
            "{}/_update/{}?refresh={}".format(
                self.Collection, self.ObjId, self.Storage.Refresh
            ),
            json=upsert_data,
        ) as resp:
            if resp.status not in {200, 201}:
                raise ConnectionError(
                    "Unexpected response code: {}".format(resp.status)
                )
            else:
                resp_json = await resp.json()
                assert (
                    resp_json["result"] == "updated" or resp_json["result"] == "created"
                ), "Creating/updating was unsuccessful"
                return self.ObjId


def serialize(v):
    if isinstance(v, datetime.datetime):
        return v.timestamp()
    else:
        return v


def get_url_list(urls):
    server_urls = []
    for url in urls:
        scheme, netloc, path = parse_url(url)

        server_urls += [
            urllib.parse.urlunparse((scheme, netloc, path, None, None, None))
            for netloc in netloc.split(",")
        ]

    return server_urls


def parse_url(url):
    parsed_url = urllib.parse.urlparse(url)
    url_path = parsed_url.path
    if not url_path.endswith("/"):
        url_path += "/"

    return parsed_url.scheme, parsed_url.netloc, url_path


def build_headers(username, password, api_key):
    # Check configurations
    if (
        username != ""
        and username is not None
        and api_key != ""
        and api_key is not None
    ):
        raise ValueError(
            "Both username and API key can't be specified. Please choose one option."
        )

    headers = {
        "Content-Type": "application/json",
    }

    # Build headers
    if username != "" and username is not None:
        auth = aiohttp.BasicAuth(username, password)
        headers["Authorization"] = auth.encode()

    elif api_key != "" and api_key is not None:
        headers["Authorization"] = "ApiKey {}".format(api_key)

    return headers


def section_has_ssl_option(config_section_name):
    """
    Checks if cert, key, cafile, capath, cadata etc. appears in section's items
    """
    for item in Config.options(config_section_name):
        if item in SSLContextBuilder.ConfigDefaults:
            return True
    return False



import logging
import aioftp
from ..abc.connection import Connection


###

L = logging.getLogger(__name__)

###


class FTPConnection(Connection):
    """
    Description:

    |

    """

    ConfigDefaults = {
        "hostname": "localhost",
        "port": 21,
    }

    def __init__(self, app, id=None, config=None):
        """
        Description:

        **Parameters**

        app : Application
                Name of the Application

        id : ID, default = None

        config : JSON, default = None

        """
        super().__init__(app, id=id, config=config)

        self.Username = self.Config.get("username")
        self.Password = self.Config.get("password")
        self.Hostname = self.Config.get("hostname")
        self.Port = self.Config.get("port")

    async def connect(self):
        """
        Description:

        :return: client

        |

        """
        client = aioftp.Client()
        await client.connect(self.Hostname, 21)
        await client.login(self.Username, self.Password)
        return client



import numpy as np
import os


class ClosedRows(object):
    def __init__(self, max_len=None):
        self.CR = set()
        if max_len is None:
            max_len = float("inf")

        self.MaxLen = max_len

    def pop(self):
        return self.CR.pop()

    def get_rows(self):
        return self.CR

    def add(self, element):
        if len(self.CR) == self.MaxLen:
            raise RuntimeError("Maximum size exceeded")

        self.CR.add(element)

    def __contains__(self, element):
        if element in self.CR:
            return True
        return False

    def serialize(self):
        return list(self.get_rows())

    def deserialize(self, data):
        self.CR = set(data)

    def __len__(self):
        return len(self.CR)

    def extend(self, start, stop):
        self.CR |= frozenset(range(start, stop))
        if len(self.CR) >= self.MaxLen:
            raise RuntimeError("Maximum size exceeded")

    def flush(self, size=None):
        self.CR = set()


class PersistentClosedRows(ClosedRows):
    def __init__(self, path, size=None, max_len=None):
        super().__init__(max_len=max_len)
        self.DType = "i1"
        self.Path = path
        if os.path.exists(self.Path):
            self.CRBit = np.memmap(self.Path, dtype=self.DType, mode="readwrite")
            for i in range(len(self.CRBit)):
                if self.CRBit[i] == 0:
                    self.CR.add(i)
        else:
            if size is None:
                raise RuntimeError("The size should correspond to array size")
            self.CR.add(0)
            self.ones(size)

    def pop(self):
        element = super().pop()
        self.CRBit[element] = 1
        return element

    def add(self, element):
        super().add(element)
        self.CRBit[element] = 0

    def extend(self, start, stop):
        cr_ = np.zeros(self.CRBit.shape[0], dtype=self.DType)
        cr_[:] = self.CRBit[:]
        cr_.resize(stop, refcheck=False)
        self.CRBit = np.memmap(self.Path, dtype=self.DType, mode="w+", shape=cr_.shape)
        self.CRBit[:] = cr_[:]
        super().extend(start, stop)

    def flush(self, size):
        super().flush(size)
        self.ones(size)

    def ones(self, size):
        self.CRBit = np.memmap(self.Path, dtype=self.DType, mode="w+", shape=(size,))
        self.CRBit[:] = 1



import abc
import copy
import time
from .. import Config


class Metric(abc.ABC):
    def __init__(self, init_values=None):
        self.Init = init_values
        self.Storage = None
        self.StaticTags = dict()

        # Expiration is relevant only to WithDynamicTagsMixIn metrics
        self.Expiration = float(Config.get("asab:metrics", "expiration"))

    def _initialize_storage(self, storage: dict):
        assert storage["type"] is None
        storage["type"] = self.__class__.__name__

        self.Storage = storage
        self.add_field(self.StaticTags)

    def add_field(self, tags):
        raise NotImplementedError(":-(")

    def flush(self, now):
        pass


class Gauge(Metric):
    def add_field(self, tags):
        field = {
            "tags": tags,
            "values": self.Init.copy() if self.Init is not None else dict(),
            "measured_at": self.App.time(),
        }
        self.Storage["fieldset"].append(field)
        self._field = field
        return field

    def set(self, name: str, value):
        """
        The function sets a value for a given value name and updates the 'measured_at' field with
        the current time.

        Args:
                name (str): The name of the value to be set.
                value: The value that you want to set for the given name in the field.
        """
        self._field["values"][name] = value
        self._field["measured_at"] = self.App.time()


class Counter(Metric):
    def add_field(self, tags):
        field = {
            "tags": tags,
            "values": self.Init.copy() if self.Init is not None else dict(),
            "actuals": self.Init.copy() if self.Init is not None else dict(),
            "measured_at": self.App.time(),
        }
        self.Storage["fieldset"].append(field)
        self._actuals = field["actuals"]
        self._field = field
        return field

    def add(self, name: str, value, init_value: dict = None):
        """
        The `add` function adds a specified value to a counter, and if the counter does not exist, it
        creates it and initializes it with an optional initial value.

        Args:
                name (str): Value name to which the `value` will be added.
                value: Value that needs to be added to the counter. It is the amount by which the counter will be incremented.
                init_value (dict): The `init_value` parameter is an optional parameter that specifies the initial value
                        of the counter if it doesn't already exist in the counter values. If `init_value` is provided, it
                        will be added to the `value` parameter and the result will be assigned as the value of the counter.
        """
        try:
            self._actuals[name] += value
        except KeyError:
            if init_value is not None:
                self._actuals[name] = init_value + value
            else:
                self._actuals[name] = value
        if not self.Storage.get("reset"):
            self._field["measured_at"] = self.App.time()

    def sub(self, name: str, value, init_value: dict = None):
        """
        The function subtracts a value from a variable and updates the 'measured_at' field.

        Args:
                name (str): The name of the variable or field that you want to subtract the value from.
                value: Value that needs to be added to the counter. It is the	amount by which the counter will be incremented.
                init_value (dict): The `init_value` parameter is an optional parameter that specifies the initial value
                        of the counter if it doesn't already exist in the counter values. If `init_value` is provided, it
                        will be added to the `value` parameter and the result will be assigned as the value of the counter.
        """

        try:
            self._actuals[name] -= value
        except KeyError:
            if init_value is not None:
                self._actuals[name] = init_value - value
            else:
                self._actuals[name] = -value
        if not self.Storage.get("reset"):
            self._field["measured_at"] = self.App.time()

    def flush(self, now):
        if self.Storage.get("reset") is True:
            self._field["measured_at"] = now
            for field in self.Storage["fieldset"]:
                field["values"] = field["actuals"]
                if self.Init is not None:
                    field["actuals"] = self.Init.copy()
                else:
                    field["actuals"] = dict()
                self._actuals = field["actuals"]
        else:
            for field in self.Storage["fieldset"]:
                field["values"] = field["actuals"].copy()


class EPSCounter(Counter):
    """
    Event per Second Counter
    Divides the count of event by a time difference between measurements.
    It effectively produces the EPS metric.
    The type of the metric is an integer (int).
    """

    def __init__(self, init_values=None):
        if init_values is not None:
            init_values = {k: int(v) for k, v in init_values.items()}
        super().__init__(init_values=init_values)
        self.LastTime = time.time()

    def flush(self, now):
        self._field["measured_at"] = now

        delta = now - self.LastTime
        if delta <= 0.0:
            return

        reset = self.Storage.get("reset")

        for field in self.Storage["fieldset"]:
            field["values"] = {k: int(v / delta) for k, v in self._actuals.items()}

            if reset is True:
                if self.Init is not None:
                    field["actuals"] = self.Init.copy()
                else:
                    field["actuals"] = dict()
                self._actuals = field["actuals"]

                self.LastTime = now


class DutyCycle(Metric):
    """
    https://en.wikipedia.org/wiki/Duty_cycle
    """

    def __init__(self, app, init_values=None):
        super().__init__()
        self.App = app
        now = self.App.time()
        self.EmptyValue = {
            "on_off": None,
            "timestamp": now,
            "off_cycle": 0.0,
            "on_cycle": 0.0,
        }

        self.Init = dict()

        if init_values is not None:
            for k, v in init_values.items():
                value = self.EmptyValue.copy()
                value["on_off"] = v
                self.Init[k] = value

    def add_field(self, tags):
        field = {
            "tags": tags,
            "actuals": self.Init.copy(),
            "values": dict(),
            "measured_at": self.App.time(),
        }
        self.Storage["fieldset"].append(field)
        self._field = field
        return field

    def set(self, name: str, on_off: bool):
        """
        The function `set` updates the state of a variable, tracking the duration of on and off cycles.

        Args:
                name (str): Name of the value that you want to set.
                on_off (bool): The `on_off` parameter is a boolean value that represents whether something is
                turned on. It is used to set the state of a particular item or feature.

        """
        now = self.App.time()
        values = self._field["actuals"].get(name)
        if values is None:
            value = self.EmptyValue.copy()
            value["on_off"] = on_off
            value["timestamp"] = now
            self._field["actuals"][name] = value
            return

        if values.get("on_off") == on_off:
            return  # No change

        d = now - values.get("timestamp")
        off_cycle = values.get("off_cycle")
        on_cycle = values.get("on_cycle")
        if on_off:
            # From off to on
            off_cycle += d
        else:
            # From on to off
            on_cycle += d

        values["on_off"] = on_off
        values["timestamp"] = now
        values["off_cycle"] = off_cycle
        values["on_cycle"] = on_cycle

    def flush(self, now):
        self._field["measured_at"] = now
        for field in self.Storage["fieldset"]:
            actuals = field.get("actuals")
            for v_name, values in actuals.items():
                d = now - values.get("timestamp")
                off_cycle = values.get("off_cycle")
                on_cycle = values.get("on_cycle")
                if values.get("on_off"):
                    on_cycle += d
                else:
                    off_cycle += d

            full_cycle = on_cycle + off_cycle
            if full_cycle > 0.0:
                field["values"][v_name] = on_cycle / full_cycle

            new_value = self.EmptyValue.copy()
            new_value["on_off"] = values.get("on_off")
            new_value["timestamp"] = now

            field["actuals"][v_name] = new_value


class AggregationCounter(Counter):
    """
    Sets value aggregated with the last one.
    Takes a function object as the `aggregator` argument.
    The aggregation function can take two arguments only.
    Maximum is used as a default aggregation function.
    """

    def __init__(self, init_values=None, aggregator=max):
        super().__init__(init_values=init_values)
        self.Aggregator = aggregator

    def set(self, name, value):
        """
        The function sets a value in a dictionary, updating it if the key already exists, and also updates a
        'measured_at' field if a certain condition is met.

        Args:
                name: Name of the value being set.
                value: Value that you want to set for the given name.
        """
        if not self.Storage.get("reset"):
            self._field["measured_at"] = self.App.time()
        try:
            self._actuals[name] = self.Aggregator(value, self._actuals[name])
        except KeyError:
            self._actuals[name] = value

    def add(self, name, value):
        raise NotImplementedError(
            "Do not use add() method with AggregationCounter. Use set() instead."
        )

    def sub(self, name, value):
        raise NotImplementedError(
            "Do not use sub() method with AggregationCounter. Use set() instead."
        )


class Histogram(Metric):
    """
    Creates cumulative histograms.
    """

    def __init__(self, buckets: list, init_values=None):
        super().__init__(init_values)
        _buckets = [float(b) for b in buckets]

        if _buckets != sorted(buckets):
            raise ValueError("Buckets not in sorted order")

        if _buckets and _buckets[-1] != float("inf"):
            _buckets.append(float("inf"))

        if len(_buckets) < 2:
            raise ValueError("Must have at least two buckets")

        self.InitBuckets = {b: dict() for b in _buckets}
        self.Count = 0
        self.Sum = 0.0
        self.InitHistogram = {"buckets": self.InitBuckets, "sum": 0.0, "count": 0}

        if self.Init:
            for value_name, value in self.Init.items():
                for upper_bound in self.InitHistogram["buckets"]:
                    if value <= upper_bound:
                        self.InitHistogram["buckets"][upper_bound][value_name] = 1
                self.InitHistogram["sum"] += value
                self.InitHistogram["count"] += 1

    def add_field(self, tags):
        field = {
            "tags": tags,
            "values": copy.deepcopy(self.InitHistogram),
            "actuals": copy.deepcopy(self.InitHistogram),
            "measured_at": self.App.time(),
        }
        self.Storage["fieldset"].append(field)
        self._actuals = field["actuals"]
        self._field = field
        return field

    def flush(self, now):
        if self.Storage.get("reset") is True:
            self._field["measured_at"] = now
            for field in self.Storage["fieldset"]:
                field["values"] = field["actuals"]
                field["actuals"] = copy.deepcopy(self.InitHistogram)
                self._actuals = field["actuals"]
        else:
            for field in self.Storage["fieldset"]:
                field["values"] = copy.deepcopy(field["actuals"])

    def set(self, value_name, value):
        """
        The function updates the values of the histogram based on the input value and value name.

        Args:
                value_name: String that represents the name of the value being set.
                value: Value that needs to be set.
        """
        if not self.Storage.get("reset"):
            self._field["measured_at"] = self.App.time()
        buckets = self._actuals["buckets"]
        summary = self._actuals["sum"]
        count = self._actuals["count"]
        for upper_bound in buckets:
            if value <= upper_bound:
                if buckets[upper_bound].get(value_name) is None:
                    buckets[upper_bound][value_name] = 1
                else:
                    buckets[upper_bound][value_name] += 1
        self._actuals["sum"] = summary + value
        self._actuals["count"] = count + 1


###


class MetricWithDynamicTags(Metric):
    def _initialize_storage(self, storage: dict):
        storage.update(
            {
                "type": self.__class__.__name__,
            }
        )
        self.Storage = storage
        if self.Init is not None:
            self.add_field(self.StaticTags.copy())

    def locate_field(self, tags):
        fieldset = self.Storage["fieldset"]

        tags = tags.copy()
        tags.update(self.StaticTags)

        # Seek for field in the fieldset using tags
        for field in fieldset:
            if field["tags"] == tags:
                return field

        # Field not found, create a new one
        field = self.add_field(tags)
        return field


class CounterWithDynamicTags(MetricWithDynamicTags):
    def add_field(self, tags):
        field = {
            "tags": tags,
            "values": self.Init.copy() if self.Init is not None else dict(),
            "actuals": self.Init.copy() if self.Init is not None else dict(),
            "expires_at": self.App.time() + self.Expiration,
            "measured_at": self.App.time(),
        }
        self.Storage["fieldset"].append(field)
        return field

    def add(self, name, value, tags):
        """
        The `add` function adds a specified value to a counter, and if the counter does not exist, it
        creates it and initializes it with an optional initial value.

        Args:
                name (str): Value name to which the `value` will be added.
                value: Value that needs to be added to the counter. It is the amount by which the counter will be incremented.
                init_value (dict): The `init_value` parameter is an optional parameter that specifies the initial value
                        of the counter if it doesn't already exist in the counter values. If `init_value` is provided, it
                        will be added to the `value` parameter and the result will be assigned as the value of the counter.
                tags (dict): Dynamic tags appliying to this value.
        """

        field = self.locate_field(tags)
        actuals = field["actuals"]
        try:
            actuals[name] += value
        except KeyError:
            actuals[name] = value

        if self.Storage.get("reset") is False:
            field["measured_at"] = self.App.time()

        field["expires_at"] = self.App.time() + self.Expiration

    def sub(self, name, value, tags):
        """
        The function subtracts a value from a variable and updates the 'measured_at' field.

        Args:
                name (str): The name of the variable or field that you want to subtract the value from.
                value: Value that needs to be added to the counter. It is the	amount by which the counter will be incremented.
                init_value (dict): The `init_value` parameter is an optional parameter that specifies the initial value
                        of the counter if it doesn't already exist in the counter values. If `init_value` is provided, it
                        will be added to the `value` parameter and the result will be assigned as the value of the counter.
                tags (dict): Dynamic tags appliying to this value.
        """

        field = self.locate_field(tags)
        actuals = field["actuals"]
        try:
            actuals[name] -= value
        except KeyError:
            actuals[name] = -value

        if self.Storage.get("reset") is False:
            field["measured_at"] = self.App.time()

        field["expires_at"] = self.App.time() + self.Expiration

    def flush(self, now):
        # Filter expired fields
        for field in self.Storage["fieldset"][::-1]:
            if field["expires_at"] < now:
                self.Storage["fieldset"].remove(field)

        if self.Storage.get("reset") is True:
            for field in self.Storage["fieldset"]:
                field["values"] = field["actuals"]
                if self.Init is not None:
                    field["actuals"] = self.Init.copy()
                else:
                    field["actuals"] = dict()
                field["measured_at"] = self.App.time()
        else:
            for field in self.Storage["fieldset"]:
                field["values"] = field["actuals"].copy()


class AggregationCounterWithDynamicTags(CounterWithDynamicTags):
    def __init__(self, init_values=None, aggregator=max):
        super().__init__(init_values=init_values)
        self.Aggregator = aggregator

    def set(self, name, value, tags):
        """
        The function sets a value for a given name and tags in a field, updates the measured_at and
        expires_at timestamps, and handles resetting the field if necessary.

        Args:
                name (str): String that represents the name of the value being set.
                value: Value that you want to set for a specific field.
                tags (dict): Dictionary of tags that are used to locate a specific field.
        """
        field = self.locate_field(tags)
        actuals = field["actuals"]
        try:
            actuals[name] = self.Aggregator(value, actuals[name])
        except KeyError:
            actuals[name] = value

        if self.Storage.get("reset") is False:
            field["measured_at"] = self.App.time()

        field["expires_at"] = self.App.time() + self.Expiration

    def add(self, name, value, tags):
        raise NotImplementedError(
            "Do not use add() method with AggregationCounter. Use set() instead."
        )

    def sub(self, name, value, tags):
        raise NotImplementedError(
            "Do not use sub() method with AggregationCounter. Use set() instead."
        )


class HistogramWithDynamicTags(MetricWithDynamicTags):
    """
    Creates cumulative histograms with dynamic tags
    """

    def __init__(self, buckets: list, init_values=None):
        super().__init__(init_values)
        _buckets = [float(b) for b in buckets]

        if _buckets != sorted(buckets):
            raise ValueError("Buckets not in sorted order")

        if _buckets and _buckets[-1] != float("inf"):
            _buckets.append(float("inf"))

        if len(_buckets) < 2:
            raise ValueError("Must have at least two buckets")

        self.InitBuckets = {b: dict() for b in _buckets}
        self.Count = 0
        self.Sum = 0.0
        self.InitHistogram = {"buckets": self.InitBuckets, "sum": 0.0, "count": 0}

        if self.Init:
            for value_name, value in self.Init.items():
                for upper_bound in self.InitHistogram["buckets"]:
                    if value <= upper_bound:
                        self.InitHistogram["buckets"][upper_bound][value_name] = 1
                self.InitHistogram["sum"] += value
                self.InitHistogram["count"] += 1

    def add_field(self, tags):
        field = {
            "tags": tags,
            "values": copy.deepcopy(self.InitHistogram),
            "actuals": copy.deepcopy(self.InitHistogram),
            "expires_at": self.App.time() + self.Expiration,
            "measured_at": self.App.time(),
        }
        self.Storage["fieldset"].append(field)
        return field

    def flush(self, now):
        # Filter expired fields
        for field in self.Storage["fieldset"][::-1]:
            if field["expires_at"] < now:
                self.Storage["fieldset"].remove(field)

        if self.Storage.get("reset") is True:
            for field in self.Storage["fieldset"]:
                field["values"] = field["actuals"]
                field["actuals"] = copy.deepcopy(self.InitHistogram)
                field["measured_at"] = self.App.time()
        else:
            for field in self.Storage["fieldset"]:
                field["values"] = copy.deepcopy(field["actuals"])

    def set(self, value_name, value, tags: dict):
        """
        The function updates the values of the histogram based on the input value and value name.

        Args:
                value_name (str): String that represents the name of the value being set.
                value: Value that needs to be set.
                tags (dict): Dynamic tags appliying to this value.
        """
        field = self.locate_field(tags)
        buckets = field.get("actuals").get("buckets")
        summary = field.get("actuals").get("sum")
        count = field.get("actuals").get("count")
        for upper_bound in buckets:
            if value <= upper_bound:
                if buckets[upper_bound].get(value_name) is None:
                    buckets[upper_bound][value_name] = 1
                else:
                    buckets[upper_bound][value_name] += 1
        field.get("actuals")["sum"] = summary + value
        field.get("actuals")["count"] = count + 1

        if self.Storage.get("reset") is False:
            field["measured_at"] = self.App.time()

        field["expires_at"] = self.App.time() + self.Expiration



import io
import asyncio
import hashlib
import typing
import logging
import functools
import os.path
import urllib.parse

import kazoo.exceptions

from .abc import LibraryProviderABC
from ..item import LibraryItem
from ...zookeeper import ZooKeeperContainer

#

L = logging.getLogger(__name__)

#


class ZooKeeperLibraryProvider(LibraryProviderABC):

    """

    Configuration variant:


    1) ZooKeeper provider is fully configured from [zookeeper] section

    .. code::

            [zookeeper]
            servers=zookeeper-1:2181,zookeeper-2:2181,zookeeper-3:2181
            path=/library

            [library]
            providers:
                    zk://



    2) ZooKeeper provider is configured by `servers` from [zookeeper] section and path from URL

    Path will be `/library`.

    .. code::

            [zookeeper]
            servers=zookeeper-1:2181,zookeeper-2:2181,zookeeper-3:2181
            path=/else

            [library]
            providers:
                    zk:///library


    2.1) ZooKeeper provider is configured by `servers` from [zookeeper] section and path from URL

    Path will be `/`, this is a special case to 2)

    .. code::

            [zookeeper]
            servers=zookeeper-1:2181,zookeeper-2:2181,zookeeper-3:2181
            path=/else

            [library]
            providers:
                    zk:///

    3) ZooKeeper provider is fully configured from URL

    .. code::

            [library]
            providers:
                    zk://zookeeper-1:2181,zookeeper-2:2181,zookeeper-3:2181/library


    4) ZooKeeper provider is configured by `servers` from [zookeeper] section and  joined `path` from [zookeeper] and
    path from URL

    Path will be `/else/library`

    .. code::

            [zookeeper]
            servers=zookeeper-1:2181,zookeeper-2:2181,zookeeper-3:2181
            path=/else

            [library]
            providers:
                    zk://./library


    If `path` from [zookeeper] section is missing, an application class name will be used
    Ex. `/BSQueryApp/library`

    """

    def __init__(self, library, path, layer):
        super().__init__(library, layer)

        url_pieces = urllib.parse.urlparse(path)

        self.FullPath = url_pieces.scheme + "://"
        self.BasePath = url_pieces.path.lstrip("/")
        while self.BasePath.endswith("/"):
            self.BasePath = self.BasePath[:-1]
        self.BasePath = "/" + self.BasePath
        if self.BasePath == "/":
            self.BasePath = ""

        if url_pieces.netloc in ["", "."]:
            # if netloc is not provided `zk:///path`, then use `zookeeper` section from config
            config_section_name = "zookeeper"
            z_url = None
        else:
            config_section_name = ""
            z_url = path

        # Initialize ZooKeeper client
        zksvc = self.App.get_service("asab.ZooKeeperService")
        self.ZookeeperContainer = ZooKeeperContainer(
            zksvc, config_section_name=config_section_name, z_path=z_url
        )
        self.Zookeeper = self.ZookeeperContainer.ZooKeeper

        if config_section_name == "zookeeper":
            self.FullPath += self.ZookeeperContainer.Config["servers"]
        else:
            self.FullPath += url_pieces.netloc

        # Handle `zk://` configuration
        if (
            z_url is None
            and url_pieces.netloc == ""
            and url_pieces.path == ""
            and self.ZookeeperContainer.Path != ""
        ):
            self.BasePath = "/" + self.ZookeeperContainer.Path

        # Handle `zk://./path` configuration
        if (
            z_url is None
            and url_pieces.netloc == "."
            and self.ZookeeperContainer.Path != ""
        ):
            self.BasePath = "/" + self.ZookeeperContainer.Path + self.BasePath

        self.FullPath += self.BasePath

        self.VersionNodePath = self.build_path("/.version.yaml")
        self.Version = None  # Will be read when a library become ready
        self.VersionWatch = None

        self.App.PubSub.subscribe(
            "ZooKeeperContainer.state/CONNECTED!", self._on_zk_connected
        )
        self.App.PubSub.subscribe("ZooKeeperContainer.state/LOST!", self._on_zk_lost)
        self.App.PubSub.subscribe(
            "ZooKeeperContainer.state/SUSPENDED!", self._on_zk_lost
        )
        self.App.PubSub.subscribe("Application.tick/60!", self._get_version_counter)

        # This will check a library for changes in subscribed folders even without version counter change.
        self.App.PubSub.subscribe("Application.tick/60!", self._on_library_changed)

        self.Subscriptions = {}

    async def finalize(self, app):
        """
        The `finalize` function is called when the application is shutting down
        """
        await self.Zookeeper._stop()

    async def _on_zk_connected(self, event_name, zkcontainer):
        """
        When the Zookeeper container is connected, set the self.Zookeeper property to the Zookeeper object.
        """
        if zkcontainer != self.ZookeeperContainer:
            return

        L.info("is connected.", struct_data={"path": self.FullPath})

        def on_version_changed(version, event):
            self.App.Loop.call_soon_threadsafe(self._check_version_counter, version)

        def install_watcher():
            return kazoo.recipe.watchers.DataWatch(
                self.Zookeeper.Client, self.VersionNodePath, on_version_changed
            )

        self.VersionWatch = await self.Zookeeper.ProactorService.execute(
            install_watcher
        )

        await self._set_ready()

    async def _on_zk_lost(self, event_name, zkcontainer):
        if zkcontainer != self.ZookeeperContainer:
            return

        await self._set_ready(ready=False)

    async def _get_version_counter(self, event_name=None):
        if self.Zookeeper is None:
            return

        version = await self.Zookeeper.get_data(self.VersionNodePath)
        self._check_version_counter(version)

    def _check_version_counter(self, version):
        # If version is `None` aka `/.version.yaml` doesn't exists, then assume version -1
        if version is not None:
            try:
                version = int(version)
            except ValueError:
                version = 1
        else:
            version = 1

        if self.Version is None:
            # Initial grab of the version
            self.Version = version
            return

        if self.Version == version:
            # The version has not changed
            return

        asyncio.create_task(self._on_library_changed())

    async def read(self, path: str) -> typing.IO:
        if self.Zookeeper is None:
            L.warning(
                "Zookeeper Client has not been established (yet). Cannot read {}".format(
                    path
                )
            )
            raise RuntimeError(
                "Zookeeper Client has not been established (yet). Not ready."
            )

        node_path = self.build_path(path)

        try:
            node_data = await self.Zookeeper.get_data(node_path)
        except kazoo.exceptions.ConnectionClosedError:
            L.warning("Zookeeper library provider is not ready")
            raise RuntimeError("Zookeeper library provider is not ready")
        except kazoo.exceptions.NoNodeError:
            return None

        # Consider adding other exceptions from Kazoo to indicate common non-critical errors

        if node_data is not None:
            return io.BytesIO(initial_bytes=node_data)
        else:
            return None

    async def list(self, path: str) -> list:
        if self.Zookeeper is None:
            L.warning(
                "Zookeeper Client has not been established (yet). Cannot list {}".format(
                    path
                )
            )
            raise RuntimeError(
                "Zookeeper Client has not been established (yet). Not ready."
            )

        node_path = self.build_path(path)

        nodes = await self.Zookeeper.get_children(node_path)
        if nodes is None:
            raise KeyError("Not '{}' found".format(node_path))

        items = []
        for node in nodes:
            # Remove any component that starts with '.'
            startswithdot = functools.reduce(
                lambda x, y: x or y.startswith("."), node.split(os.path.sep), False
            )
            if startswithdot:
                continue

            if (
                "." in node
            ):  # We detect files in zookeeper by presence of the dot in the filename,
                fname = path + node
                ftype = "item"
            else:
                fname = path + node + "/"
                ftype = "dir"

            items.append(
                LibraryItem(
                    name=fname,
                    type=ftype,
                    layer=self.Layer,
                    providers=[self],
                )
            )

        return items

    def build_path(self, path):
        """
        It takes a path in the library and transforms in into a path within Zookeeper.
        It does also series of sanity checks (asserts).

        IMPORTANT: If you encounter asserting failure, don't remove assert.
        It means that your code is incorrect.
        """
        assert path[:1] == "/"
        if path != "/":
            node_path = self.BasePath + path
        else:
            node_path = self.BasePath

        # Zookeeper path should not have forward slash at the end of path
        node_path = node_path.rstrip("/")

        assert (
            "//" not in node_path
        ), "Directory path cannot contain double slashes (//). Example format: /library/Templates/"
        assert (
            node_path[0] == "/"
        ), "Directory path must start with a forward slash (/). For example: /library/Templates/"

        return node_path

    async def subscribe(self, path):
        path = self.BasePath + path
        self.Subscriptions[path] = await self._get_directory_hash(path)

    async def _get_directory_hash(self, path):
        def recursive_traversal(path, digest):
            if not self.Zookeeper.Client.exists(path):
                return

            children = self.Zookeeper.Client.get_children(path)
            for child in children:
                if path != "/":
                    child_path = "{}/{}".format(path, child)
                else:
                    child_path = "/{}".format(child)
                zstat = self.Zookeeper.Client.exists(child_path)
                digest.update(
                    "{}\n{}\n".format(child_path, zstat.version).encode("utf-8")
                )
                recursive_traversal(child_path, digest)

        digest = hashlib.sha1()
        await self.Zookeeper.ProactorService.execute(recursive_traversal, path, digest)
        return digest.digest()

    async def _on_library_changed(self, event_name=None):
        for path, digest in self.Subscriptions.items():
            newdigest = await self._get_directory_hash(path)
            if newdigest != digest:
                self.Subscriptions[path] = newdigest
                self.App.PubSub.publish("Library.change!", self, path)



from ..abc.source import Source
from ..abc.sink import Sink

import copy
import asyncio
import bspump.jupyter

import nest_asyncio

nest_asyncio.apply()


class TestSource(Source):
    def __init__(self, *args, **kwargs):
        kwargs["test_events"] = kwargs.get("test_events", {})
        self.test_events = kwargs["test_events"]
        del kwargs["test_events"]
        super().__init__(*args, **kwargs)

    def start(self, loop):
        """
        Starts the :meth:`Pipeline <bspump.Pipeline()>` through the _main method, but if main method is implemented
        it starts the coroutine using main method instead.

        **Parameters**

        loop : ?
                        Contains the coroutines.


        """
        if self.Task is not None:
            return

        async def _main():
            """
            Description:

            :return:
            """
            print(f"\nRunning tests for pipeline {self.Pipeline.Id}.")
            # if test_events is a dict we will iterate over its items
            # if it is a list of tuples we iterate over them
            if isinstance(self.test_events, dict):
                test_events = self.test_events.items()
            else:
                test_events = self.test_events
            for event, tests in test_events:
                tests["outputs"] = []
                tests["input"] = str(event)
                bspump.jupyter.bitswan_test_probes.clear()
                bspump.jupyter.bitswan_test_probes.update(tests.get("probes", {}))
                print(f"\n    ┌ Testing event:        {event}")
                all_tasks_before = asyncio.all_tasks()
                await self.process(event, context=tests)
                all_tasks_after = asyncio.all_tasks()
                other_tasks = [
                    task for task in all_tasks_after if task not in all_tasks_before
                ]

                # Wait for all other tasks to complete
                if other_tasks:
                    await asyncio.wait(other_tasks)
                print(f"    └ Outputs:              {tests['outputs']}", end="")
                if tests.get("expect") and tests["outputs"] != tests["expect"]:
                    print(" \033[91m✘\033[0m")
                    print(
                        f"    ! \033[91mTest failed\033[0m Expected: {tests['expect']}\n"
                    )
                    if not self.Pipeline.App.Watch:
                        exit(1)
                if tests.get("inspect"):
                    inspected = tests["inspect"][0](tests["outputs"])
                    expected = tests["inspect"][1]
                    if not inspected == expected:
                        print(" \033[91m✘\033[0m")
                        print(
                            f"    ! \033[91mInspect failed. Got {inspected} expected {expected}\033[0m"
                        )
                        if not self.Pipeline.App.Watch:
                            exit(1)
                if tests.get("inspect_in"):
                    inspected = tests["inspect_in"][0](tests["outputs"])
                    expected = tests["inspect_in"][1]
                    if expected not in inspected:
                        print(" \033[91m✘\033[0m")
                        print(
                            f'    ! \033[91mInspect failed. Got "{inspected}" expected that to contain "{expected}"\033[0m'
                        )
                        if not self.Pipeline.App.Watch:
                            exit(1)

                print(" \033[92m✔\033[0m")
            print(f"\n\033[92mAll tests passed for {self.Pipeline.Id}.\033[0m\n")
            bspump.jupyter.bitswan_tested_pipelines.add(self.Pipeline.Id)
            if bspump.jupyter.bitswan_tested_pipelines == set(
                self.Pipeline.App.PumpService.Pipelines
            ):
                if not self.Pipeline.App.Watch:
                    exit()

        self.Pipeline._evaluate_ready()
        asyncio.get_event_loop().run_until_complete(_main())


class TestSink(Sink):
    def process(self, context, event):
        context["outputs"].append(copy.deepcopy(event))



import os
import logging
from ..abc.sink import Sink


L = logging.getLogger(__file__)


class FileBlockSink(Sink):
    """
    Description:

    ** Config Defaults **

    path : ''

    mode : wb

    flags : O_CREAT

    """

    ConfigDefaults = {
        "path": "",
        "mode": "wb",
        "flags": "O_CREAT",
    }

    OFlagDict = {"O_CREAT": os.O_CREAT, "O_EXCL": os.O_EXCL}

    def __init__(self, app, pipeline, id=None, config=None):
        """

        **Parameters**

        app : Application
                Name of the Application

        pipeline : Pipeline
                Name of the Pipeline.

        id : ID, default = None
                ID

        config : JSON, default = None
                Configuration file with additional information.

        """
        super().__init__(app, pipeline, id=id, config=config)

        self._oflags = 0
        for flag in self.Config["flags"].split(","):
            flag = flag.strip()
            try:
                self._oflags |= self.OFlagDict[flag]
            except KeyError:
                L.warning("Unknown oflag '{}'".format(flag))

    def get_file_name(self, context, event):
        """
        Override this method to gain control over output file name.

        **Parameters**

        context :

        event : any type
                a single unit of information that is propagated through the pipeline

        :return: config path

        |

        """
        return self.Config["path"]

    def process(self, context, event):
        """
        Opens a file.

        **Parameters**

        context :

        event : any type
                a single unit of information that is propagated through the pipeline

        """
        fname = self.get_file_name(context, event)

        fd = os.open(fname, os.O_WRONLY | self._oflags)
        with os.fdopen(fd, self.Config["mode"]) as fo:
            fo.write(event)



from .connection import PostgreSQLConnection
from .sink import PostgreSQLSink
from .logicalreplicationsource import PostgreSQLLogicalReplicationSource
from .lookup import PostgreSQLLookup
from .source import PostgreSQLSource

__all__ = (
    "PostgreSQLConnection",
    "PostgreSQLSink",
    "PostgreSQLLogicalReplicationSource",
    "PostgreSQLLookup",
    "PostgreSQLSource",
)



from aiohttp.web import Request

from bspump.http.web.components.field import Field
from bspump.http.web.template_env import env


class FloatField(Field):
    def inner_html(self, default=0, readonly=False):
        if not default:
            default = 0.0

        template = env.get_template("number-field.html")
        return template.render(
            default=default,
            default_input_props=self.default_input_props,
            default_classes=self.default_classes,
        )

    def clean(self, data, request: Request | None = None):
        if type(data.get(self.name)) == str:
            data[self.name] = float(data.get(self.name, 0))



from ..abc.sink import Sink


class PostgreSQLSink(Sink):
    ConfigDefaults = {
        "query": "",
        "data": "",
    }

    def __init__(self, app, pipeline, connection, id=None, config=None):
        super().__init__(app, pipeline, id=id, config=config)

        self._connection = pipeline.locate_connection(app, connection)
        self._query = self.Config["query"]
        self._data_keys = self.Config["data"].split(",")

        app.PubSub.subscribe("PostgreSQLConnection.pause!", self._connection_throttle)
        app.PubSub.subscribe("PostgreSQLConnection.unpause!", self._connection_throttle)

    def process(self, context, event):
        # Prepare data
        data = tuple(event.get(x) for x in self._data_keys)

        # Consume
        self._connection.consume(self._query, data)

    def _connection_throttle(self, event_name, connection):
        if connection != self._connection:
            return

        if event_name == "PostgreSQLConnection.pause!":
            self.Pipeline.throttle(self, True)
        elif event_name == "PostgreSQLConnection.unpause!":
            self.Pipeline.throttle(self, False)
        else:
            raise RuntimeError("Unexpected event name '{}'".format(event_name))



import aiohttp.web


class ValidationError(Exception):
    """
    Request cannot be processed because it does not match expected schema
    """

    # TODO: Inherit from aiohttp.web.HTTPBadRequest
    pass


class NotAuthenticatedError(aiohttp.web.HTTPUnauthorized):
    """
    Request could not be authenticated
    """

    def __init__(self):
        # TODO: Optionally include "error", "realm" etc. (https://www.rfc-editor.org/rfc/rfc6750#section-3)
        super().__init__(headers={"WWW-Authenticate": "Bearer scope=openid"})


class AccessDeniedError(aiohttp.web.HTTPForbidden):
    """
    Authenticated subject does not have the rights to access requested resource
    """

    pass


class Conflict(Exception):
    """
    Request cannot be satisfied because it would introduce a state that violates some uniqueness requirement
    """

    # TODO: Inherit from aiohttp.web.HTTPConflict
    # TODO: Handle when the value of `key` or `value` is actually `None`
    def __init__(self, message=None, *args, key=None, value=None):
        self.Key = key
        self.Value = value

        if message is None:
            if key is not None:
                if value is not None:
                    message = "Conflict in field {}: {}".format(repr(key), repr(value))
                else:
                    message = "Conflict in field {}".format(repr(key))
            elif value is not None:
                message = value

        if message is None:
            super().__init__(*args)
        else:
            super().__init__(message, *args)



import os

import bspump.declarative
import bspump.unittest


class TestDeclarativeDict(bspump.unittest.TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.Builder = bspump.declarative.ExpressionBuilder(self.App)
        self.Optimizer = bspump.declarative.ExpressionOptimizer(self.App)

    def load(self, decl_fname):
        basedir = os.path.dirname(__file__)
        with open(os.path.join(basedir, decl_fname), "r") as f:
            return self.Builder.parse(f.read())[0]

    def test_dict_01(self):
        event = {
            "string": "STRING",
            "integer": 15,
        }

        decl = self.load("./test_dict_01.yaml")

        res = decl({}, event)
        self.assertEqual(res, {"item1": "STRING", "item2": 15})

    def test_dict_02(self):
        context = {
            "JSON": {
                "EndTime": "Tomorrow",
                "DeviceProperties": [
                    {"Name": "OS", "Value": "Windows 10"},
                    {"Name": "BrowserType", "Value": "Chrome"},
                    {"Name": "IsCompliantAndManaged", "Value": "False"},
                ],
            }
        }

        decl = self.load("./test_dict_02.yaml")

        res = self.Optimizer.optimize(decl)(context, {})

        self.assertEqual(
            res,
            {
                "o365.audit.EndTime": "Tomorrow",
                "o365.audit.DeviceProperties.OS": "Windows 10",
                "o365.audit.DeviceProperties.BrowserType": "Chrome",
                "o365.audit.DeviceProperties.IsCompliantAndManaged": "False",
            },
        )



import re
import logging

import aiohttp
import aiohttp.web

import typing

from .accesslog import AccessLogger
from ..config import Configurable
from ..tls import SSLContextBuilder
from .service import WebService
from ..application import Application

#

L = logging.getLogger(__name__)

#


class WebContainer(Configurable):
    """
    Configurable object that serves as a backend for `asab.WebService`.
    It contains everything needed for the web server existence, namely all the configuration and the server Application object.
    """

    ConfigDefaults = {
        "listen": "0.0.0.0 8080",  # Can be multiline
        "backlog": 128,
        "rootdir": "",
        "servertokens": "full",  # Controls whether 'Server' response header field is included ('full') or faked 'prod' ()
        "cors": "",
        "cors_preflight_paths": "/openidconnect/*, /test/*",
        "body_max_size": 1024**2,  # Client’s maximum body size in a request, in bytes
    }

    def __init__(
        self,
        websvc: WebService,
        config_section_name: str,
        config: typing.Optional[dict] = None,
    ):
        super().__init__(config_section_name=config_section_name, config=config)

        self.Addresses = None  # The address is available only after `WebContainer.started!` PubSub message is published.
        self.BackLog = int(self.Config.get("backlog"))
        self.CORS = self.Config.get("cors")

        servertokens = self.Config.get("servertokens")
        if servertokens == "prod":
            # Because we cannot remove token completely
            self.ServerTokens = "asab"
        else:
            from .. import __version__

            self.ServerTokens = (
                aiohttp.web_response.SERVER_SOFTWARE + " asab/" + __version__
            )

        # Parse listen address(es), can be multiline configuration item
        ls = self.Config.get("listen")
        self._listen = []
        for line in ls.split("\n"):
            line = line.strip()
            if len(line) == 0:
                continue

            if " " in line:
                line = re.split(r"\s+", line)
            else:
                # This line allows the (obsolete) format of IPv4 with ':'
                # such as "0.0.0.0:8001"
                line = re.split(r"[:\s]", line, 1)

            if all([c in "0123456789" for c in line[0]]):
                # If the first item is a number, consider that a port number
                addr = ["0.0.0.0", "::"]  # We want to listen on IPv4 and IPv6
                port = line.pop(0).strip()
                port = int(port)
            else:
                # First item is a port, a second is an IP address of the network interface to listen to
                addr = line.pop(0).strip()
                port = line.pop(0).strip()
                port = int(port)
            ssl_context = None

            for param in line:
                if param.startswith("ssl:"):
                    # Dedicated section for SSL
                    ssl_context = SSLContextBuilder(param, config=self.Config).build()
                    # SSL parameters are included in the current config section
                elif param.startswith("ssl"):
                    ssl_context = SSLContextBuilder(
                        "<none>", config=self.Config
                    ).build()
                else:
                    raise RuntimeError(
                        "Unknown listen parameter in section [{}]: {}".format(
                            config_section_name, param
                        )
                    )

            if isinstance(addr, list):
                for a in addr:
                    self._listen.append((a, port, ssl_context))
            else:
                self._listen.append((addr, port, ssl_context))

        if len(self._listen) == 0:
            L.warning("Missing configuration.")

        client_max_size = int(self.Config.get("body_max_size"))
        self.WebApp: aiohttp.web.Application = aiohttp.web.Application(
            client_max_size=client_max_size
        )
        """
		The Web Application object. See [aiohttp documentation](https://docs.aiohttp.org/en/stable/web_reference.html?highlight=Application#application) for the details.

		It is a *dict-like* object, so you can use it for sharing data globally by storing arbitrary properties for later access from a handler.

		Attributes:
			WebApp["app"] (asab.Application): Reference to the ASAB Application.

			WebApp["rootdir"] (asab.web.staticdir.StaticDirProvider): Reference to the root path specified by `rootdir` configuration.
		"""
        self.WebApp.on_response_prepare.append(self._on_prepare_response)
        self.WebApp["app"] = websvc.App

        rootdir = self.Config.get("rootdir")
        if len(rootdir) > 0:
            from .staticdir import StaticDirProvider

            self.WebApp["rootdir"] = StaticDirProvider(
                self.WebApp, root="/", path=rootdir
            )

        access_log = logging.getLogger(__name__[: __name__.rfind(".")] + ".al")
        access_log.App = websvc.App

        self.WebAppRunner = aiohttp.web.AppRunner(
            self.WebApp,
            handle_signals=False,
            access_log=access_log,
            access_log_class=AccessLogger,
        )

        websvc._register_container(self, config_section_name)

        if self.CORS != "":
            preflight_str = (
                self.Config["cors_preflight_paths"]
                .strip("\n")
                .replace("*", "{tail:.*}")
            )
            preflight_paths = re.split(r"[,\s]+", preflight_str, re.MULTILINE)
            self.add_preflight_handlers(preflight_paths)

    async def _start(self, app: Application):
        await self.WebAppRunner.setup()

        for addr, port, ssl_context in self._listen:
            site = aiohttp.web.TCPSite(
                self.WebAppRunner,
                host=addr,
                port=port,
                backlog=self.BackLog,
                ssl_context=ssl_context,
            )
            try:
                await site.start()
            except OSError as err:
                L.error(
                    "Cannot start web server: {}".format(err),
                    struct_data={"address": addr, "port": port},
                )

            if isinstance(site, aiohttp.web_runner.TCPSite):
                for address in site._runner.addresses:
                    if self.Addresses is None:
                        self.Addresses = []
                    self.Addresses.append(address)

        self.WebApp["app"].PubSub.publish("WebContainer.started!", self)

    async def _stop(self, app: Application):
        self.WebApp["app"].PubSub.publish("WebContainer.stopped!", self)
        await self.WebAppRunner.cleanup()

    def add_preflight_handlers(self, preflight_paths: typing.List[str]):
        """
        Add handlers with preflight resources.

        Preflight requests are sent by the browser, for some cross domain request (custom header etc.).
        Browser sends preflight request first.
        It is request on the same endpoint as app demanded request, but of **OPTIONS** method.
        Only when satisfactory response is returned, browser proceeds with sending original request.
        Use `cors_preflight_paths` option to specify all paths and path prefixes (separated by comma) for which you
        want to allow **OPTIONS** method for preflight requests.

        Args:
                preflight_paths (list[str]): List of routes that will be provided with **OPTIONS** handler.
        """
        for path in preflight_paths:
            self.WebApp.router.add_route("OPTIONS", path, self._preflight_handler)

    async def _preflight_handler(self, request):
        return aiohttp.web.HTTPNoContent(
            headers={
                "Access-Control-Allow-Origin": request.headers.get("Origin", "*"),
                "Access-Control-Allow-Methods": "GET, POST, OPTIONS",
                "Access-Control-Allow-Headers": "X-PINGOTHER, Content-Type, Authorization",
                "Access-Control-Allow-Credentials": "true",
                "Access-Control-Max-Age": "86400",
            }
        )

    async def _on_prepare_response(self, request, response):
        response.headers["Server"] = self.ServerTokens

        if self.CORS != "":
            # TODO: Be more precise about "allow origin" header
            response.headers["Access-Control-Allow-Origin"] = "*"
            response.headers[
                "Access-Control-Allow-Methods"
            ] = "GET, POST, DELETE, PUT, PATCH, OPTIONS"

    def get_ports(self) -> typing.List[str]:
        """
        Return list of available ports.

        Returns:
                (list[str]) List of ports.
        """
        ports = []
        for addr, port, ssl_context in self._listen:
            ports.append(port)
        return ports



import bspump.unittest
import bspump.common


class TestNullSink(bspump.unittest.ProcessorTestCase):
    def test_null(self):
        events = [
            (None, "Don't let this out!"),
        ]

        self.set_up_processor(bspump.common.NullSink)

        output = self.execute(events)

        self.assertEqual([event for context, event in output], [])



import logging
import os

import bspump.asab.proactor as proactor

from bspump.abc.lookupprovider import LookupBatchProviderABC

###

L = logging.getLogger(__name__)

###


class FileBatchLookupProvider(LookupBatchProviderABC):
    """
    Loads lookup data from a file on local filesystem.

    |

    """

    def __init__(self, lookup, url, id=None, config=None):
        super().__init__(lookup, url, id, config)
        self.App.add_module(proactor.Module)
        self.ProactorService = self.App.get_service("asab.ProactorService")

    async def load(self):
        """
        Description:

        :returns: result

        |

        """
        result = await self.ProactorService.execute(
            self.load_on_thread,
        )
        return result

    def load_on_thread(self):
        """
        Description:

        |

        """
        if not os.path.isfile(self.URL):
            L.warning("Source '{}' is not a file".format(self.URL))
            return None
        if not os.access(self.URL, os.R_OK):
            L.warning("Insufficient permissions: '{}'".format(self.URL))
            return None
        try:
            with open(self.URL, "rb") as f:
                data = f.read()
            return data
        except Exception as e:
            L.warning("Failed to read content of file '{}': {}".format(self.URL, e))
        return None

    async def save(self, data):
        """

        **Parameters**

        data :

        """
        dirname = os.path.dirname(self.URL)
        if not os.path.isdir(dirname):
            os.makedirs(dirname)
        with open(self.URL, "wb") as fo:
            fo.write(data)



import logging

from .analyzer import Analyzer
from ..matrix.sessionmatrix import SessionMatrix, PersistentSessionMatrix

###

L = logging.getLogger(__name__)

###


class SessionAnalyzer(Analyzer):
    """
    This is the analyzer for events with multiple different dimensions.

    `SessionAnalyzer` operates over the `SessionMatrix` object.
    `column_formats` is an array, each element contains the letter from the table + number:

            +------------+------------------+
            | Name       | Definition       |
            +============+==================+
            | 'b'        | Byte             |
            +------------+------------------+
            | 'i'        | Signed integer   |
            +------------+------------------+
            | 'u'        | Unsigned integer |
            +------------+------------------+
            | 'f'        | Floating point   |
            +------------+------------------+
            | 'c'        | Complex floating |
            |            | point            |
            +------------+------------------+
            | 'S'        | String           |
            +------------+------------------+
            | 'U'        | Unicode string   |
            +------------+------------------+
            | 'V'        | Raw data         |
            +------------+------------------+

    Example: 'i8' stands for int64.
    Important! However it is possible to use all these letters, it is recommeded to use only 'i' for integers, 'f' for
    floats, 'U' for strings. Anything else might cause problems in serialization.
    It is possible to create a matrix with elements of specified format. The tuple with number of dimensions should
    stand before the letter.
    Example: '(6, 3)i8' will create the matrix with n rows, 6 columns and 3 third dimensions with integer elements.
    `column_names` is an array with names of each column, with the same length as `column_formats`.
    `matrix_id` is an id of `SessionMatrix` object defined alternatively.

    """

    def __init__(
        self,
        app,
        pipeline,
        matrix_id=None,
        dtype="float_",
        analyze_on_clock=False,
        persistent=False,
        id=None,
        config=None,
    ):
        """
        Description:

        **Parameters**

        app : Application
                Name of the Application.

        pipeline : Pipeline
                Name of the Pipeline

        matrix_id : str, default = None

        dtype : str, default = 'float_'

        analyze_on_clock : bool, default = False

        persistent : bool, defualt = False

        id : str, defualt = None

        config : JSON, default = None
                configuration file with additional information.

        """
        super().__init__(
            app, pipeline, analyze_on_clock=analyze_on_clock, id=id, config=config
        )
        svc = app.get_service("bspump.PumpService")
        if matrix_id is None:
            s_id = self.Id + "Matrix"
            if persistent:
                self.Sessions = PersistentSessionMatrix(
                    app, dtype, id=s_id, config=config
                )
            else:
                self.Sessions = SessionMatrix(app, dtype, id=s_id, config=config)
            svc.add_matrix(self.Sessions)
        else:
            self.Sessions = svc.locate_matrix(matrix_id)



import bspump.asab
import asyncio
from .trigger import Trigger


class PeriodicTrigger(Trigger):
    def __init__(self, app, interval=None, id=None):
        """
        Interval is in seconds, can be a float or int.
        """

        super().__init__(app, id)
        self.Timer = bspump.asab.Timer(app, self.on_timer, autorestart=True)
        self.Timer.start(interval)

        async def run_on_ready():
            for pipeline in app.PumpService.Pipelines.values():
                while not pipeline.is_ready():
                    await asyncio.sleep(0.1)
            await self.Timer.Handler()

        asyncio.ensure_future(run_on_ready())

    async def on_timer(self):
        self.fire()

    @classmethod
    def construct(cls, app, definition: dict):
        id = definition.get("id")
        interval = definition.get("args", {}).get("interval", 1.0)
        return cls(app, id=id, interval=interval)



import pyjq
from bspump.abc.processor import Processor


class JQProcessor(Processor):
    """
    JQProcessor is a processor that uses JQ query language to transform data.
    """

    def __init__(self, app, pipeline, id=None, config=None, query=None):
        super().__init__(app, pipeline, id=id, config=config)
        self.Query = query or "{}".format(self.Config["query"])

    def process(self, context, event):
        return pyjq.first(self.Query, event)


# MockProcessor inherits from JQProcessor to avoid dependency on BSPump
class MockProcessor(JQProcessor):
    def __init__(self, app, pipeline, *args, **kwargs):
        self.Query = "Mock Query"


def test_jq():
    # Arrange
    processor = MockProcessor(None, None, None)
    processor.Query = '{"hi": (.foo * 5) }'
    event = {"foo": 3}

    # Act
    result_event = processor.process(None, event)

    # Assert
    assert result_event == {"hi": 15}


def test_multiple_jq():
    # Arrange
    processor = MockProcessor(None, None, None)
    processor.Query = (
        '{"result1": (.foo * 5), "result2": (.foo * 10), "result3": (.foo * 15) }'
    )
    event = {"foo": 3}

    # Act
    result_event = processor.process(None, event)

    # Assert
    assert result_event == {"result1": 15, "result2": 30, "result3": 45}


def test_concat_jq():
    # Arrange
    processor = MockProcessor(None, None, None)
    processor.Query = (
        '{"result1": (.value2 + .value3), "result2": (.value1 * .value4) }'
    )
    event = {"value1": 3, "value2": "hi", "value3": "!", "value4": 5}

    # Act
    result_event = processor.process(None, event)

    # Assert
    assert result_event == {"result1": "hi!", "result2": 15}


def test_typecast_jq():
    # Arrange
    processor = MockProcessor(None, None, None)
    processor.Query = '{"result1": (.value1 + (.value2|tonumber)) }'
    event = {"value1": 3, "value2": "2"}

    # Act
    result_event = processor.process(None, event)

    # Assert
    assert result_event == {"result1": 5}



import logging
import bspump.asab as asab

#

L = logging.getLogger(__name__)

#


class Module(asab.Module):
    def __init__(self, app):
        super().__init__(app)
        sttype = asab.Config.get("asab:storage", "type", fallback=None)

        if sttype == "inmemory":
            from .inmemory import StorageService

            self.Service = StorageService(app, "asab.StorageService")

        elif sttype == "mongodb":
            from .mongodb import StorageService

            self.Service = StorageService(app, "asab.StorageService")

        elif sttype == "elasticsearch":
            from .elasticsearch import StorageService

            self.Service = StorageService(app, "asab.StorageService")

        elif sttype is None:
            L.critical("Missing configuration for [asab:storage] type.")
            raise SystemExit("Exit due to a critical configuration error.")

        else:
            L.critical(
                "Unknown configuration type '{}' in [asab:storage].".format(sttype)
            )
            raise SystemExit("Exit due to a critical configuration error.")



import logging
import numpy as np
import os
from .matrix import Matrix, PersistentMatrix

#
L = logging.getLogger(__name__)
#


class GeoMatrix(Matrix):
    """
    Matrix, specific for `GeoAnalyzer`.
    `bbox` is the dictionary with `max_lat`, `min_lat`, `max_lon` and `min_lon`
    for corner gps-coordinates.
    `GeoMatrix` is 2d projection of real-world coordinates on a plane.

    """

    ConfigDefaults = {
        "max_lat": 71.26,  # Europe endpoints
        "min_lat": 23.33,
        "min_lon": -10.10,
        "max_lon": 40.6,
    }

    def __init__(
        self, app, dtype="float_", bbox=None, resolution=5, id=None, config=None
    ):
        if bbox is None:
            bbox = {
                "min_lat": float(self.ConfigDefaults["min_lat"]),
                "max_lat": float(self.ConfigDefaults["max_lat"]),
                "min_lon": float(self.ConfigDefaults["min_lon"]),
                "max_lon": float(self.ConfigDefaults["max_lon"]),
            }

        self.Bbox = bbox
        self.Resolution = resolution
        self._update_matrix_dimensions()

        super().__init__(app, dtype=dtype, id=id, config=config)

    def zeros(self):
        self.Array = np.zeros([self.MapHeight, self.MapWidth], dtype=self.DType)

    def is_in_boundaries(self, lat, lon):
        """
        Check, if coordinates are within the bbox coordinates.
        """
        if (lat >= self.Bbox["max_lat"]) or (lat <= self.Bbox["min_lat"]):
            return False

        if (lon >= self.Bbox["max_lon"]) or (lon <= self.Bbox["min_lon"]):
            return False

        return True

    def _update_matrix_dimensions(self):
        """
        Calculation of MapHeight and MapWidth.
        """
        self.SizeWidth = self.get_gps_distance(
            self.Bbox["min_lat"],
            self.Bbox["min_lon"],
            self.Bbox["min_lat"],
            self.Bbox["max_lon"],
        )
        self.SizeHeight = self.get_gps_distance(
            self.Bbox["min_lat"],
            self.Bbox["min_lon"],
            self.Bbox["max_lat"],
            self.Bbox["min_lon"],
        )
        self.MapHeight = int(np.ceil(self.SizeHeight / self.Resolution))
        self.MapWidth = int(np.ceil(self.SizeWidth / self.Resolution))

        # Correction
        self.SizeWidth = int(self.MapWidth * self.Resolution)
        self.SizeHeight = int(self.MapHeight * self.Resolution)

    def degrees_to_radians(self, degrees):
        return degrees * np.pi / 180

    def get_gps_distance(self, lat1, lon1, lat2, lon2):
        """
        Calculation of distance between 2 gps-coordinates in km.
        """
        R = 6371
        dLat = self.degrees_to_radians(lat2 - lat1)
        dLon = self.degrees_to_radians(lon2 - lon1)
        lat1 = self.degrees_to_radians(lat1)
        lat2 = self.degrees_to_radians(lat2)
        a = np.sin(dLat / 2) * np.sin(dLat / 2) + np.sin(dLon / 2) * np.sin(
            dLon / 2
        ) * np.cos(lat1) * np.cos(lat2)
        c = 2 * np.arctan2(np.sqrt(a), np.sqrt(1 - a))
        return R * c

    def project_equirectangular(self, lat, lon):
        """
        Converts latitude and longitude into row and column indexes.
        """
        column = (lon - self.Bbox["min_lon"]) * (
            (self.MapWidth - 1) / (self.Bbox["max_lon"] - self.Bbox["min_lon"])
        )
        row = ((lat * (-1)) + self.Bbox["max_lat"]) * (
            (self.MapHeight - 1) / (self.Bbox["max_lat"] - self.Bbox["min_lat"])
        )
        return int(row), int(column)

    def inverse_equirectangular(self, row, column):
        """
        Converts row and column into latitude and longitude.
        """
        row += 0.5
        column += 0.5

        lat = -(
            (row / (self.MapHeight - 1) * (self.Bbox["max_lat"] - self.Bbox["min_lat"]))
            - self.Bbox["max_lat"]
        )
        lon = (
            column * (self.Bbox["max_lon"] - self.Bbox["min_lon"]) / (self.MapWidth - 1)
            + self.Bbox["min_lon"]
        )

        return lat, lon


class PersistentGeoMatrix(PersistentMatrix):
    """
    Matrix, specific for `GeoAnalyzer`.
    `bbox` is the dictionary with `max_lat`, `min_lat`, `max_lon` and `min_lon`
    for corner gps-coordinates.
    `GeoMatrix` is 2d projection of real-world coordinates on a plane with pointers
    to `IdsToMembers`, where objects can be kept.

    """

    ConfigDefaults = {
        "max_lat": 71.26,  # Europe endpoints
        "min_lat": 23.33,
        "min_lon": -10.10,
        "max_lon": 40.6,
    }

    def __init__(
        self, app, dtype="float_", bbox=None, resolution=5, id=None, config=None
    ):
        if bbox is None:
            bbox = {
                "min_lat": float(self.ConfigDefaults["min_lat"]),
                "max_lat": float(self.ConfigDefaults["max_lat"]),
                "min_lon": float(self.ConfigDefaults["min_lon"]),
                "max_lon": float(self.ConfigDefaults["max_lon"]),
            }

        self.Bbox = bbox
        self.Resolution = resolution
        self._update_matrix_dimensions()

        super().__init__(app, dtype=dtype, id=id, config=config)

    def zeros(self):
        self.create_path()
        if os.path.exists(self.ArrayPath):
            self.Array = np.memmap(self.ArrayPath, dtype=self.DType, mode="readwrite")
            self.Array = self.Array.reshape(self.reshape(self.Array.shape))
        else:
            array = np.zeros([self.MapHeight, self.MapWidth], dtype=self.DType)
            self.Array = np.memmap(
                self.ArrayPath, dtype=self.DType, mode="w+", shape=array.shape
            )

    def reshape(self, shape):
        return [self.MapHeight, self.MapWidth]

    def is_in_boundaries(self, lat, lon):
        """
        Check, if coordinates are within the bbox coordinates.
        """
        if (lat >= self.Bbox["max_lat"]) or (lat <= self.Bbox["min_lat"]):
            return False

        if (lon >= self.Bbox["max_lon"]) or (lon <= self.Bbox["min_lon"]):
            return False

        return True

    def _update_matrix_dimensions(self):
        """
        Calculation of MapHeight and MapWidth.
        """
        self.SizeWidth = self.get_gps_distance(
            self.Bbox["min_lat"],
            self.Bbox["min_lon"],
            self.Bbox["min_lat"],
            self.Bbox["max_lon"],
        )
        self.SizeHeight = self.get_gps_distance(
            self.Bbox["min_lat"],
            self.Bbox["min_lon"],
            self.Bbox["max_lat"],
            self.Bbox["min_lon"],
        )
        self.MapHeight = int(np.ceil(self.SizeHeight / self.Resolution))
        self.MapWidth = int(np.ceil(self.SizeWidth / self.Resolution))

        # Correction
        self.SizeWidth = int(self.MapWidth * self.Resolution)
        self.SizeHeight = int(self.MapHeight * self.Resolution)

    def degrees_to_radians(self, degrees):
        return degrees * np.pi / 180

    def get_gps_distance(self, lat1, lon1, lat2, lon2):
        """
        Calculation of distance between 2 gps-coordinates in km.
        """
        R = 6371
        dLat = self.degrees_to_radians(lat2 - lat1)
        dLon = self.degrees_to_radians(lon2 - lon1)
        lat1 = self.degrees_to_radians(lat1)
        lat2 = self.degrees_to_radians(lat2)
        a = np.sin(dLat / 2) * np.sin(dLat / 2) + np.sin(dLon / 2) * np.sin(
            dLon / 2
        ) * np.cos(lat1) * np.cos(lat2)
        c = 2 * np.arctan2(np.sqrt(a), np.sqrt(1 - a))
        return R * c

    def project_equirectangular(self, lat, lon):
        """
        Converts latitude and longitude into row and column indexes.
        """
        column = (lon - self.Bbox["min_lon"]) * (
            (self.MapWidth - 1) / (self.Bbox["max_lon"] - self.Bbox["min_lon"])
        )
        row = ((lat * (-1)) + self.Bbox["max_lat"]) * (
            (self.MapHeight - 1) / (self.Bbox["max_lat"] - self.Bbox["min_lat"])
        )
        return int(row), int(column)

    def inverse_equirectangular(self, row, column):
        """
        Converts row and column into latitude and longitude.
        """
        row += 0.5
        column += 0.5

        lat = -(
            (row / (self.MapHeight - 1) * (self.Bbox["max_lat"] - self.Bbox["min_lat"]))
            - self.Bbox["max_lat"]
        )
        lon = (
            column * (self.Bbox["max_lon"] - self.Bbox["min_lon"]) / (self.MapWidth - 1)
            + self.Bbox["min_lon"]
        )

        return lat, lon



import re
import abc
import socket
import logging
import asyncio

import aiohttp
from bspump.asab import Service, Config, Configurable
import dataclasses

#

L = logging.getLogger(__name__)

#


@dataclasses.dataclass
class Alert:
    source: str
    alert_cls: str
    alert_id: str
    title: str
    data: dict
    detail: str = ""
    exception: Exception = None


class AlertProviderABC(Configurable, abc.ABC):
    ConfigDefaults = {}

    def __init__(self, config_section_name):
        super().__init__(config_section_name=config_section_name)

    async def initialize(self, app):
        pass

    async def finalize(self, app):
        pass

    @abc.abstractmethod
    def trigger(self, alert: Alert):
        pass


class AlertAsyncProviderABC(AlertProviderABC):
    def __init__(self, config_section_name):
        super().__init__(config_section_name=config_section_name)
        self.Queue = asyncio.Queue()
        self.MainTask = None

    async def initialize(self, app):
        self._start_main_task()

    async def finalize(self, app):
        if self.MainTask is not None:
            mt = self.MainTask
            self.MainTask = None
            mt.cancel()

    def trigger(self, alert: Alert):
        self.Queue.put_nowait(alert)

    def _start_main_task(self):
        assert self.MainTask is None
        self.MainTask = asyncio.ensure_future(self._main())
        self.MainTask.add_done_callback(self._main_done)

    def _main_done(self, x):
        if self.MainTask is None:
            return

        self.MainTask.result()

        self.MainTask = None
        self._start_main_task()

    @abc.abstractmethod
    async def _main(self):
        pass


class AlertHTTPProviderABC(AlertAsyncProviderABC):
    ConfigDefaults = {
        "url": "",
    }

    def __init__(self, config_section_name):
        super().__init__(config_section_name=config_section_name)
        self.URL = self.Config["url"]


class OpsGenieAlertProvider(AlertHTTPProviderABC):
    ConfigDefaults = {
        # US: https://api.opsgenie.com
        # EU: https://api.eu.opsgenie.com
        "url": "https://api.eu.opsgenie.com",
        # See https://docs.opsgenie.com/docs/authentication
        # E.g. `eb243592-faa2-4ba2-a551q-1afdf565c889`
        "api_key": "",
        # Coma separated tags to be added to the request
        "tags": "",
    }

    def __init__(self, config_section_name, *args, **kwargs):
        super().__init__(config_section_name=config_section_name)
        self.APIKey = self.Config["api_key"]
        self.Tags = re.split(r"[,\s]+", self.Config["tags"], re.MULTILINE)
        self.Hostname = socket.gethostname()

    async def _main(self):
        while True:
            a = await self.Queue.get()

            headers = {"Authorization": "GenieKey {}".format(self.APIKey)}

            create_alert = {
                "message": a.title,
                "note": a.detail,
                "alias": "{}:{}:{}".format(a.source, a.alert_cls, a.alert_id),
                "tags": self.Tags,
                "details": {
                    "source": a.source,
                    "class": a.alert_cls,
                    "id": a.alert_id,
                },
                "entity": a.source,
                "source": self.Hostname,
            }

            if a.data:
                create_alert["details"].update(a.data)

            async with aiohttp.ClientSession(headers=headers) as session:
                async with session.post(
                    self.URL + "/v2/alerts", json=create_alert
                ) as resp:
                    if resp.status != 202:
                        text = await resp.text()
                        L.warning(
                            "Failed to create the alert ({}):\n'{}'".format(
                                resp.status, text
                            )
                        )
                    else:
                        await resp.text()


class PagerDutyAlertProvider(AlertHTTPProviderABC):
    ConfigDefaults = {
        "url": "https://events.pagerduty.com",
        # Your api key generated by PagerDuty
        "api_key": "",
        # Integration key (or routing_key) from a Service directory
        # Choose "Use our API directly" and "Events API v2"
        "integration_key": "",
    }

    def __init__(self, config_section_name, *args, **kwargs):
        super().__init__(config_section_name=config_section_name)
        self.APIKey = self.Config["api_key"]
        self.IntegrationKey = self.Config["integration_key"]

    async def _main(self):
        while True:
            a = await self.Queue.get()

            headers = {"Authorization": "Token token={}".format(self.APIKey)}

            create_alert = {
                "event_action": "trigger",
                "routing_key": self.IntegrationKey,
                "dedup_key": "{}:{}:{}".format(a.source, a.alert_cls, a.alert_id),
                "client": "Asab Alert Service",
                "payload": {
                    "summary": a.title,
                    "severity": "warning",
                    "source": a.source,
                    "group": a.alert_cls,
                    "custom_details": {
                        "source": a.source,
                        "class": a.alert_cls,
                        "id": a.alert_id,
                    },
                },
            }

            if a.data:
                create_alert["payload"]["custom_details"].update(a.data)

            async with aiohttp.ClientSession(headers=headers) as session:
                async with session.post(
                    self.URL + "/v2/enqueue", json=create_alert
                ) as resp:
                    if resp.status != 202:
                        text = await resp.text()
                        L.warning(
                            "Failed to create the alert ({}):\n'{}'".format(
                                resp.status, text
                            )
                        )
                    else:
                        await resp.text()


class SentryAlertProvider(AlertAsyncProviderABC):
    def __init__(self, config_section_name="", application=None):
        super().__init__(config_section_name=config_section_name)
        import bspump.asab.sentry

        self.SentryService = bspump.asab.sentry.SentryService(application)

    async def _main(self):
        while True:
            a = await self.Queue.get()
            self.SentryService.set_tags(
                {"source": a.source, "class": a.alert_cls, "id": a.alert_id}
            )
            self.SentryService.capture_exception(a.exception)

    async def finalize(self, app):
        await super().finalize(app)


class AlertService(Service):
    def __init__(self, app, service_name="asab.AlertService"):
        super().__init__(app, service_name)
        self.Providers = []

        for section in Config.sections():
            if not section.startswith("asab:alert:"):
                continue

            provider_cls = {
                "asab:alert:opsgenie": OpsGenieAlertProvider,
                "asab:alert:pagerduty": PagerDutyAlertProvider,
                "asab:alert:sentry": SentryAlertProvider,
            }.get(section)
            if provider_cls is None:
                L.warning("Unknwn alert provider: {}".format(section))
                continue

            self.Providers.append(
                provider_cls(config_section_name=section, application=app)
            )

    async def initialize(self, app):
        await asyncio.gather(*[p.initialize(app) for p in self.Providers])

    async def finalize(self, app):
        await asyncio.gather(*[p.finalize(app) for p in self.Providers])

    def trigger(
        self,
        alert: Alert,
    ):
        for p in self.Providers:
            p.trigger(alert)



# IPC = InterProcess Communication

from .datagram import DatagramSource, DatagramSink

from .stream_client_sink import StreamClientSink
from .stream_server_source import StreamServerSource

# Backward compatibility
StreamSource = StreamServerSource
StreamSink = StreamClientSink

__all__ = (
    "DatagramSource",
    "DatagramSink",
    "StreamServerSource",
    "StreamClientSink",
    "StreamSink",
    "StreamSource",
)



import datetime
import pytz

from bspump.asab import Config

from ...abc import Expression
from ..value.valueexpr import VALUE


class DATETIME_PARSE(Expression):
    """
    Returns date/time in human readable format.
    The date is created from `datetime`, which by default is current UTC time.

    Format example: "%Y-%m-%d %H:%M:%S"
    """

    Attributes = {
        "Value": ["*"],  # TODO: This ...
        "Format": ["*"],  # TODO: This ...
        "Timezone": ["*"],  # TODO: This ...
    }

    Category = "Date/Time"

    def __init__(self, app, *, arg_what, arg_format, arg_flags="", arg_timezone=None):
        super().__init__(app)
        if not isinstance(arg_what, Expression):
            self.Value = VALUE(app, value=arg_what)
        else:
            self.Value = arg_what

        if not isinstance(arg_format, Expression):
            self.Format = VALUE(app, value=arg_format)
        else:
            self.Format = arg_format

        self.SetCurrentYear = "Y" in arg_flags

        if arg_timezone is None:
            timezone_from_config = Config["declarations"]["timezone"]

            if len(timezone_from_config) == 0:
                self.Timezone = None

            else:
                self.Timezone = pytz.timezone(timezone_from_config)

        else:
            self.Timezone = pytz.timezone(arg_timezone)

    def __call__(self, context, event, *args, **kwargs):
        fmt = self.Format(context, event, *args, **kwargs)
        value = self.Value(context, event, *args, **kwargs)

        if isinstance(value, int) or isinstance(value, float):
            value = datetime.datetime.utcfromtimestamp(value)

        try:
            if fmt == "RFC3339":
                dt = datetime.datetime.strptime(value, "%Y-%m-%dT%H:%M:%S.%fZ")
            else:
                dt = datetime.datetime.strptime(value, fmt)
        except ValueError:
            return None

        if self.SetCurrentYear:
            dt = dt.replace(year=datetime.datetime.utcnow().year)

        if self.Timezone is not None:
            dt = self.Timezone.localize(dt)
        else:
            if dt.tzinfo is None:
                # Naive datatime is considered as UTC
                dt = dt.replace(tzinfo=datetime.timezone.utc)
            else:
                # Timezone aware localtime is converted to UTC
                dt = dt.astimezone(datetime.timezone.utc)

        return dt.timestamp()



import logging
import os

import fastavro
import fastavro.validation
from ..avro import loader
from bspump.abc.sink import Sink

L = logging.getLogger(__name__)


class AvroSink(Sink):
    """
    Apache Avro™ is a data serialization system.

    Avro relies on schemas defined with JSON.
    Schemas are composed of primitive types null, boolean, int, long, float, double, bytes,
    and string and complex types (record, enum, array, map, union, and fixed).
    You can learn more about Avro from their documentation - https://avro.apache.org/docs/current/

    Avro schema example:
            {
                    "namespace": "example.avro",
                    "type": "record",
                    "name": "User",
                    "fields": [
                            {"name": "name", "type": "string"},
                            {"name": "favorite_number",  "type": ["int", "null"]},
                            {"name": "favorite_color", "type": ["string", "null"]}
                    ]
            }

    File Types:
    .avro - Avro Serialized Data
    .avsc - Avro Schema

    """

    ConfigDefaults = {
        "file_name_template": "./sink{index}.avro",
        "events_per_file": 1000,
        "events_per_chunk": 100,
        "rollover_mechanism": "none",  # or chunks
    }

    def __init__(self, app, pipeline, id=None, config=None):
        super().__init__(app, pipeline, id=id, config=config)

        self.SchemaFile = self.Config.get("schema_file")
        if self.SchemaFile is None:
            raise FileNotFoundError("Avro schema is not defined")

        if self.SchemaFile is not None:
            self.Schema = loader.load_avro_schema(self.Config)

        self.RolloverMechanism = self.Config.get("rollover_mechanism")
        self.FileNameTemplate = self.Config["file_name_template"]
        self.Events = []  # list of events
        self.EventsPerFile = self.Config["events_per_file"]
        self.ChunkSize = self.Config["events_per_chunk"]
        self._filemode = "wb"

        if self.RolloverMechanism == "chunks":
            if self.EventsPerFile % self.ChunkSize != 0:
                self.ChunksPerFile = int(round(self.EventsPerFile / self.ChunkSize))
                L.warning(
                    "events_per_file % events_per_chunk needs to be zero. events_per_file was rounded"
                )
            elif self.EventsPerFile < self.ChunkSize:
                self.ChunksPerFile = 1
                L.warning(
                    "events_per_file needs to be equal or greater than events_per_chunk, events_per_file changed to events_per_chunk value"
                )
            elif self.EventsPerFile >= self.ChunkSize:
                self.ChunksPerFile = int(self.EventsPerFile / self.ChunkSize)
            else:
                L.warning(
                    "invalid events_per_file - {} or events_per_chunk - {} ".format(
                        self.EventsPerFile, self.ChunkSize
                    )
                )
                self.ChunksPerFile = 1

            self.Chunks = 0
            self.Index = 0
        else:
            if self.RolloverMechanism != "none":
                L.warning(
                    "Unknown rollover mechanism {} - defaulting to none".format(
                        self.RolloverMechanism
                    )
                )

            self.RolloverMechanism = "none"
            self.Index = None
            self.Chunks = None

    def build_filename(self, postfix=""):
        return (
            self.FileNameTemplate.format(
                index=("{:04d}".format(self.Index)) if self.Index is not None else ""
            )
            + postfix
        )

    def flush(self):
        if len(self.Events) != 0:
            if self.SchemaFile is not None:
                is_valid = fastavro.validation.validate_many(self.Events, self.Schema)
                if is_valid is True:
                    with open(self.build_filename("-open"), self._filemode) as out:
                        fastavro.writer(out, self.Schema, self.Events)
                    if self._filemode == "wb":
                        self._filemode = "a+b"

            self.Events = []

    def process(self, context, event):
        self.Events.append(event)

        if self.RolloverMechanism == "chunks":
            if len(self.Events) >= self.EventsPerFile:
                self.Chunks = self.Chunks + 1
                if self.Chunks >= self.ChunksPerFile:
                    self.rotate()
        self.flush()

    def rotate(self, new_filename=None):
        """
        Call this to close the currently open file.
        """

        self.flush()

        current_fname = self.build_filename("-open")
        os.rename(current_fname, current_fname[:-5])

        if self.RolloverMechanism == "chunks":
            self.Index = self.Index + 1
        self._filemode = "wb"



import logging

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes

from ..abc.processor import Processor

#

L = logging.getLogger(__name__)

#


class HashingBaseProcessor(Processor):
    ConfigDefaults = {
        "algorithm": "sha256",
        "digest_size": 64,
    }

    def __init__(self, app, pipeline, id=None, config=None):
        super().__init__(app, pipeline, id, config)

        self.Backend = default_backend()

        algorithm = self.Config["algorithm"].upper()

        if algorithm == "SHA224":
            self.Algorithm = hashes.SHA224()
        elif algorithm == "SHA256":
            self.Algorithm = hashes.SHA256()
        elif algorithm == "SHA384":
            self.Algorithm = hashes.SHA384()
        elif algorithm == "SHA512":
            self.Algorithm = hashes.SHA512()
        elif algorithm == "SHA1":
            self.Algorithm = hashes.SHA1()
        elif algorithm == "MD5":
            self.Algorithm = hashes.MD5()
        elif algorithm == "BLAKE2B":
            digest_size = int(self.Config["digest_size"])
            self.Algorithm = hashes.BLAKE2b(digest_size)
        elif algorithm == "BLAKE2S":
            digest_size = int(self.Config["digest_size"])
            self.Algorithm = hashes.BLAKE2s(digest_size)
        else:
            L.error("Unknown hashing algorithm '{}'".format(self.Config["algorithm"]))
            raise RuntimeError(
                "Unknown hashing algorithm '{}'".format(self.Config["algorithm"])
            )


class HashingProcessor(HashingBaseProcessor):
    """
    Create hash of the event.
    """

    def process(self, context, event):
        digest = hashes.Hash(self.Algorithm, self.Backend)
        digest.update(event)
        return digest.finalize()


class CoHashingProcessor(HashingBaseProcessor):
    """
    Put hash of the event into a context.
    """

    def process(self, context, event):
        digest = hashes.Hash(self.Algorithm, self.Backend)
        digest.update(event)
        context["hash"] = digest.finalize()
        return event



import logging

from ..abc.service import Service
from .. import Config
from .. import metrics

from .metrics import WebRequestsMetrics


L = logging.getLogger(__name__)


class WebService(Service):
    """
    Service for running and easy manipulation of the web server.
    It is used for registering and running the web container as well as initialization of web request metrics.

    It should be used together with [`asab.web.WebContainer`](#asab.web.WebContainer) object that handles the web configuration.

    Examples:

    ```python
    from asab.web import Module
    self.add_module(Module)
    web_service = self.get_service("asab.WebService")
    container = asab.web.WebContainer(
            websvc=web_service,
            config_section_name='my:web',
            config={"listen": "0.0.0.0:8080"}
    )
    ```
    """

    ConfigSectionAliases = ["asab:web"]

    def __init__(self, app, service_name):
        super().__init__(app, service_name)

        # Web service is dependent on Metrics service
        if Config.getboolean("asab:metrics", "web_requests_metrics", fallback=False):
            app.add_module(metrics.Module)
            self.MetricsService = app.get_service("asab.MetricsService")
            self.WebRequestsMetrics = WebRequestsMetrics(self.MetricsService)
        self.Containers = {}

    async def finalize(self, app):
        for containers in self.Containers.values():
            await containers._stop(app)

    def _register_container(self, container, config_section_name: str):
        self.Containers[config_section_name] = container
        self.App.TaskService.schedule(container._start(self.App))

    @property
    def WebApp(self):
        """
        An obsolete property only for maintaining backward compatibility. Please use `asab.web.WebContainer.WebApp` instead.
        """
        return self.WebContainer.WebApp

    @property
    def WebContainer(self):
        """
        An obsolete property only for maintaining backward compatibility. Please use `asab.web.WebContainer` instead.
        """
        config_section = "web"

        # The WebContainer should be configured in the config section [web]
        if config_section not in Config.sections():
            # If there is no [web] section, try other aliases for backwards compatibility
            for alias in self.ConfigSectionAliases:
                if alias in Config.sections():
                    config_section = alias
                    L.warning(
                        "Using obsolete config section [{}]. Preferred section name is [web]. ".format(
                            alias
                        )
                    )
                    break
            else:
                raise RuntimeError("No [web] section configured.")

        try:
            return self.Containers[config_section]
        except KeyError:
            from .container import WebContainer

            return WebContainer(self, config_section)



from ..abc.processor import Processor

from .generalanomaly import GeneralAnomaly


class AnomalyManager(Processor):
    """
    AnomalyManager stores symptoms of anomalies to create an anomaly object based on key
    dimensions within the symptom

    Dimensions are stored as part of individual symptoms inside the anomaly object.
    Other dimensions than key dimensions are optional and serve only to differentiate
    the symptoms from one another.

    The final anomaly objects, which can be obtained from the AnomalyStorage, then contain
    the following fields (where @timestamp is the timestamp of the first symptom):

    {
            "@timestamp": UNIX_TIMESTAMP,
            "ts_end": UNIX_TIMESTAMP,
            "type": STRING,
            "status": STRING,
            "symptoms": ARRAY,
            "key_dimension1": ...
            "key_dimension2": ...
    }

    When the symptom contains status set to closed and the anomaly is specified to finish by that flag,
    the anomaly "ts_end" is copied from this symptom and the anomaly object moved from "open" to "closed"
    inside the anomaly storage.
    """

    def __init__(
        self,
        app,
        pipeline,
        anomaly_storage,
        anomaly_classes=list(),
        id=None,
        config=None,
    ):
        super().__init__(app, pipeline, id, config)

        self.AnomalyStorage = anomaly_storage
        self.AnomalyClasses = anomaly_classes

        metrics_service = app.get_service("asab.MetricsService")
        self.AnomalyManagerCounter = metrics_service.create_counter(
            "anomaly.manager",
            init_values={
                "timestamp.miss": 0,
                "key.miss": 0,
                "key.dimensions.miss": 0,
                "type.miss": 0,
                "symptom.miss": 0,
            },
        )

    def create_anomaly(self, key_dimensions, timestamp_started, anomaly_type):
        selected_anomaly_class = GeneralAnomaly
        for anomaly_class in self.AnomalyClasses:
            if anomaly_class.TYPE == anomaly_type:
                selected_anomaly_class = anomaly_class
                break

        anomaly = selected_anomaly_class()
        anomaly["@timestamp"] = timestamp_started
        anomaly["type"] = anomaly_type
        anomaly["status"] = "open"
        anomaly["symptoms"] = []

        # Store all key dimensions in top level
        for dim_name, dim_value in key_dimensions.items():
            anomaly[dim_name] = dim_value

        return anomaly

    def process(self, context, event):
        # Obtain timestamp
        timestamp = event.get("@timestamp")
        if timestamp is None:
            self.AnomalyManagerCounter.add("timestamp.miss", 1)
            return None

        # Obtain key
        key = event.get("key")
        if key is None:
            self.AnomalyManagerCounter.add("key.miss", 1)
            return None

        # Obtain key dimensions
        key_dimensions = event.get("key_dimensions")
        if key_dimensions is None:
            self.AnomalyManagerCounter.add("key.dimensions.miss", 1)
            return None

        # Obtain type
        anomaly_type = event.get("type")
        if anomaly_type is None:
            self.AnomalyManagerCounter.add("type.miss", 1)
            return None

        # Obtain symptom data
        symptom = event.get("symptom")
        if symptom is None:
            self.AnomalyManagerCounter.add("symptom.miss", 1)
            return None

        # Store the open anomaly in the storage
        if key not in self.AnomalyStorage["open"]:
            anomaly = self.create_anomaly(key_dimensions, timestamp, anomaly_type)
            self.AnomalyStorage["open"][key] = anomaly

        # Append symptom to the anomaly
        self.AnomalyStorage["open"][key]["symptoms"].append(symptom)

        return event



from .service import SentryService
from ..config import Config

Config.add_defaults(
    {
        "sentry": {
            "data_source_name": "",
            "environment": "not specified",
            "traces_sample_rate": 1.0,
        },
        "sentry:logging": {"breadcrumbs": "info", "events": "error"},
    }
)

__all__ = ["SentryService"]



from croniter import croniter
from datetime import datetime
from .trigger import Trigger


class CronTrigger(Trigger):
    """
    Trigger that uses crontab syntax for specifying execution frequency.

    CronTrigger allows you to schedule pipeline execution using standard
    cron expressions. It calculates the next execution time based on the
    cron string and fires when that time is reached.

    Cron syntax:
        * * * * *
        | | | | |
        | | | | +-- Day of week (0-7, 0 and 7 are Sunday)
        | | | +---- Month (1-12)
        | | +------ Day of month (1-31)
        | +-------- Hour (0-23)
        +---------- Minute (0-59)

    Examples:
        "0 9 * * *"     # Daily at 9:00 AM
        "*/15 * * * *"  # Every 15 minutes
        "0 0 1 * *"     # First day of every month at midnight
        "0 9 * * 1"     # Every Monday at 9:00 AM

    Warning: Always use timezone-aware datetime objects for init_time
    to avoid timezone-related issues.
    """

    def __init__(self, app, cron_string, init_time, id=None):
        super().__init__(app, id=id)
        self.cron_string = cron_string
        self.init_time = init_time
        self.next_trigger_time = self.get_new_time(cron_string, init_time)

        app.PubSub.subscribe("Application.tick!", self.on_timer)

    async def on_timer(self, event_type="simulated"):
        """
        Method that is called on every tick of the application.
        """
        if datetime.now() > self.next_trigger_time:
            # get new time for next trigger
            self.next_trigger_time = self.get_new_time(
                self.cron_string, self.next_trigger_time
            )
            self.fire()
        else:
            return

    def pause(self, pause=True):
        """
        Pauses the trigger
        """
        super().pause(pause)
        if not pause:
            self.Loop.call_soon(self.on_tick)

    def get_new_time(self, cron_string, time):
        """
        Calculates new time for next trigger
        """
        iterable = croniter(cron_string, time)
        return iterable.get_next(datetime)

    @classmethod
    def construct(cls, app, definition: dict):
        id = definition.get("id")
        interval = definition.get("args", {}).get("cron_string")
        if interval is None:
            raise RuntimeError("CronTrigger needs interval to be defined")
        init_time = datetime.now()
        return cls(app, id=id, cron_string=interval, init_time=init_time)



import paho.mqtt.client as mqtt
import logging

from ..abc.connection import Connection

L = logging.getLogger(__name__)


class MQTTConnection(Connection):
    ConfigDefaults = {
        "broker": "localhost:1883",
        "keepalive": 60,
        "username": "",
        "password": "",
    }

    def __init__(self, app, id=None, config=None):
        super().__init__(app, id=id, config=config)
        self._client = mqtt.Client(transport="websockets")
        broker = self.Config["broker"]
        self._host, self._port = broker.split(":")

        if self._port == "443":
            # communication between public broker
            self._client.tls_set(cert_reqs=mqtt.ssl.CERT_REQUIRED)
        else:
            # internal communication between broker
            self._client.tls_set(cert_reqs=mqtt.ssl.CERT_NONE)

        if self.Config["username"] != "" and self.Config["password"] != "":
            self._client.username_pw_set(
                self.Config["username"], self.Config["password"]
            )

        self._client.connect(self._host, int(self._port), int(self.Config["keepalive"]))

        self._dispatcher = MQTTMessageDispatcher()

        self._client.on_connect = self.on_connect
        self._client.on_message = self._dispatcher.dispatch
        self.Connected = False
        self._client.loop_start()

    def on_connect(self, client, userdata, flags, rc, properties=None):
        L.info(f"Connected to MQTT broker: {self._host}:{self._port}")
        self.Connected = True

    def register_handler(self, topic, handler):
        self._dispatcher.register(topic, handler)

    def subscribe_topic(self, topic, qos=0):
        self._client.subscribe(topic, qos)

    def publish_to_topic(self, topic, payload, qos=0, retain=False):
        self._client.publish(topic, payload, qos, retain)


class MQTTMessageDispatcher:
    def __init__(self):
        self.handlers = {}

    def register(self, topic, handler):
        self.handlers[topic] = handler

    def dispatch(self, client, userdata, message):
        topic = message.topic
        if topic in self.handlers:
            self.handlers[topic](client, userdata, message)
        else:
            L.warning(f"No handler for topic {topic}")



import datetime
import typing
import motor.motor_asyncio
import pymongo
import bson

import bspump.asab as asab
from .exceptions import DuplicateError
from .service import StorageServiceABC
from .upsertor import UpsertorABC

asab.Config.add_defaults(
    {
        "asab:storage": {
            "mongodb_uri": "mongodb://localhost:27017",
            "mongodb_database": "asabdb",
        }
    }
)


class StorageService(StorageServiceABC):
    """
    StorageService for MongoDB. Depends on `pymongo` and `motor`.
    """

    def __init__(self, app, service_name, config_section_name="asab:storage"):
        super().__init__(app, service_name)
        self.Client = motor.motor_asyncio.AsyncIOMotorClient(
            asab.Config.get(config_section_name, "mongodb_uri")
        )

        self.Database = self.Client.get_database(
            asab.Config.get(config_section_name, "mongodb_database"),
            codec_options=bson.codec_options.CodecOptions(
                tz_aware=True, tzinfo=datetime.timezone.utc
            ),
        )
        assert self.Database is not None

    def upsertor(self, collection: str, obj_id=None, version=0):
        return MongoDBUpsertor(self, collection, obj_id, version)

    async def get(self, collection: str, obj_id, decrypt=None) -> dict:
        coll = self.Database[collection]
        ret = await coll.find_one({"_id": obj_id})
        if ret is None:
            raise KeyError("NOT-FOUND")
        if decrypt is not None:
            for field in decrypt:
                if field in ret:
                    ret[field] = self.aes_decrypt(ret[field])
        return ret

    async def get_by(self, collection: str, key: str, value, decrypt=None) -> dict:
        coll = self.Database[collection]
        ret = await coll.find_one({key: value})
        if ret is None:
            raise KeyError("NOT-FOUND")
        if decrypt is not None:
            for field in decrypt:
                if field in ret:
                    ret[field] = self.aes_decrypt(ret[field])
        return ret

    async def collection(
        self, collection: str
    ) -> motor.motor_asyncio.AsyncIOMotorCollection:
        """
        Get collection. Useful for custom operations.

        Args:
                collection: Collection to get.

        Returns:
                `AsyncIOMotorCollection` object connected to the queried database.

        Examples:

                >>> coll = await storage.collection("test-collection")
                >>> cursor = coll.find({})
                >>> while await cursor.fetch_next:
                ... 	obj = cursor.next_object()
                ... 	pprint.pprint(obj)

        """

        return self.Database[collection]

    async def delete(self, collection: str, obj_id):
        coll = self.Database[collection]
        ret = await coll.find_one_and_delete({"_id": obj_id})
        if ret is None:
            raise KeyError("NOT-FOUND")
        return ret["_id"]


class MongoDBUpsertor(UpsertorABC):
    @classmethod
    def generate_id(cls):
        return bson.objectid.ObjectId()

    async def execute(
        self,
        custom_data: typing.Optional[dict] = None,
        event_type: typing.Optional[str] = None,
    ):
        id_name = self.get_id_name()
        addobj = {}

        if len(self.ModSet) > 0:
            addobj["$set"] = self.ModSet

        if len(self.ModInc) > 0:
            addobj["$inc"] = self.ModInc

        if len(self.ModPush) > 0:
            addobj["$push"] = {k: {"$each": v} for k, v in self.ModPush.items()}

        if len(self.ModUnset) > 0:
            addobj["$unset"] = {k: "" for k in self.ModUnset}

        filtr = {}

        if self.ObjId is not None:
            filtr[id_name] = self.ObjId
        else:
            # We are going to insert a new object without explicit Id
            assert (self.Version == 0) or (self.Version is None)

        if self.Version is not None:
            filtr["_v"] = int(self.Version)

        # First wave (adding stuff)
        if len(addobj) > 0:
            coll = self.Storage.Database[self.Collection]
            try:
                ret = await coll.find_one_and_update(
                    filtr,
                    update=addobj,
                    upsert=True
                    if (self.Version == 0) or (self.Version is None)
                    else False,
                    return_document=pymongo.collection.ReturnDocument.AFTER,
                )
            except pymongo.errors.DuplicateKeyError as e:
                if hasattr(e, "details"):
                    raise DuplicateError(
                        "Duplicate key error: {}".format(e),
                        self.ObjId,
                        key_value=e.details.get("keyValue"),
                    )
                else:
                    raise DuplicateError(
                        "Duplicate key error: {}".format(e), self.ObjId
                    )

            if ret is None:
                # Object might have been changed in the meantime
                raise KeyError("NOT-FOUND")

            self.ObjId = ret[id_name]

        # for k, v in self.ModPull.items():
        # 	o = obj.pop(k, None)
        # 	if o is None: o = list()
        # 	for x in v:
        # 		try:
        # 			o.remove(x)
        # 		except ValueError:
        # 			pass
        # 	obj[k] = o

        if self.Storage.WebhookURIs is not None:
            webhook_data = {
                "collection": self.Collection,
            }

            if custom_data is not None:
                webhook_data["custom"] = custom_data

            if event_type is not None:
                webhook_data["event_type"] = event_type

            # Add upsetor data; do not include fields that start with "__"
            upsertor_data = {
                "id_field_name": id_name,
                "id": self.ObjId,
                "_v": int(self.Version),
            }
            if len(self.ModSet) > 0:
                upsertor_data["set"] = {
                    k: v for k, v in self.ModSet.items() if not k.startswith("__")
                }
            if len(self.ModInc) > 0:
                upsertor_data["inc"] = {
                    k: v for k, v in self.ModInc.items() if not k.startswith("__")
                }
            if len(self.ModPush) > 0:
                upsertor_data["push"] = {
                    k: v for k, v in self.ModPush.items() if not k.startswith("__")
                }
            if len(self.ModUnset) > 0:
                upsertor_data["unset"] = {
                    k: v for k, v in self.ModUnset.items() if not k.startswith("__")
                }
            webhook_data["upsertor"] = upsertor_data

            await self.webhook(webhook_data)

        return self.ObjId



import logging
from ..abc import Service
from ..log import LOG_NOTICE

#

L = logging.getLogger(__name__)

#


class NativeMetrics(Service):
    """
    This service is responsible for reading native metrics.
    There are:
    * memory metrics
    """

    def __init__(self, app, metrics_svc):
        self.MemoryGauge = metrics_svc.create_gauge("memory")

        # Injecting logging metrics into MetricsHandler and MetricsHandler into Root Logger
        self.MetricsLoggingHandler = MetricsLoggingHandler()
        self.MetricsLoggingHandler.LogCounter = metrics_svc.create_counter(
            "logs",
            init_values={"warnings": 0, "errors": 0, "critical": 0},
            help="Counts WARNING, ERROR and CRITICAL logs per minute.",
        )
        logging.root.addHandler(self.MetricsLoggingHandler)

        app.PubSub.subscribe("Metrics.flush!", self._on_flushing_event)
        self._on_flushing_event()

    def _on_flushing_event(self, event_name=None):
        try:
            with open("/proc/self/status", "r") as file:
                proc_status = file.read()

                for proc_status_line in proc_status.replace("\t", "").split("\n"):
                    # Vm - virtual memory, other metrics need to be evaluated and added
                    if proc_status_line.startswith("Vm"):
                        proc_status_info = proc_status_line.split(" ")
                        try:
                            self.MemoryGauge.set(
                                proc_status_info[0][:-1],
                                int(proc_status_info[-2]) * 1024,
                            )
                        except ValueError:
                            pass

        except FileNotFoundError:
            pass
            # L.warning("File '/proc/self/status' was not found, skipping reading native metrics.")


class MetricsLoggingHandler(logging.Handler):
    def emit(self, record):
        level = record.levelno
        if level <= LOG_NOTICE:
            return
        elif level <= logging.WARNING:
            self.LogCounter.add("warnings", 1)
        elif level <= logging.ERROR:
            self.LogCounter.add("errors", 1)
        elif level <= logging.CRITICAL:
            self.LogCounter.add("critical", 1)
